from django.db import models

# Create your models here.
from django.db.models import BooleanField, ImageField, TextField


class Product(models.Model):
    photo = ImageField(upload_to='products')

    class Meta:
        abstract = True


class Blind(Product):
    name = TextField()
    child_safe = BooleanField(default=False)

    def __str__(self):
        return self.name



from django.http import HttpResponse
from django.test import TestCase


class TestResponseAssumptions(TestCase):

    def test_headers_present_in_http_response(self):
        """Verify that HttpResponse has a headers or _headers attribute, which we use and Mock in our tests."""
        django_response = HttpResponse()
        self.assertTrue(hasattr(django_response, '_headers') or hasattr(django_response, 'headers'))



# Generated by Django 2.2.6 on 2019-10-26 12:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('silk', '0006_fix_request_prof_file_blank'),
    ]

    operations = [
        migrations.AddField(
            model_name='sqlquery',
            name='identifier',
            field=models.IntegerField(default=-1),
        ),
    ]



from django.apps import apps as proj_apps
from django.test import TestCase

from silk.apps import SilkAppConfig


class TestAppConfig(TestCase):
    """
    Test if correct AppConfig class is loaded by Django.
    """

    def test_app_config_loaded(self):
        silk_app_config = proj_apps.get_app_config("silk")
        self.assertIsInstance(silk_app_config, SilkAppConfig)



import calendar
import random
from datetime import datetime, timedelta
from itertools import groupby
from math import floor

import pytz
from django.test import TestCase
from django.utils import timezone

from silk import models
from silk.request_filters import (
    AfterDateFilter,
    BeforeDateFilter,
    FunctionNameFilter,
    MethodFilter,
    NameFilter,
    NumQueriesFilter,
    OverallTimeFilter,
    PathFilter,
    SecondsFilter,
    StatusCodeFilter,
    TimeSpentOnQueriesFilter,
    ViewNameFilter,
)

from .test_lib.mock_suite import MockSuite
from .util import delete_all_models

mock_suite = MockSuite()


class TestRequestFilters(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def _time_stamp(self, dt):
        return calendar.timegm(dt.utctimetuple())

    def test_seconds_filter(self):
        requests = [mock_suite.mock_request() for _ in range(0, 10)]
        n = 0
        for r in requests:
            r.start_time = timezone.now() - timedelta(seconds=n)
            r.save()
            n += 1
        requests = models.Request.objects.filter(SecondsFilter(5))
        for r in requests:
            dt = r.start_time
            seconds = self._time_stamp(timezone.now()) - self._time_stamp(dt)
            self.assertTrue(seconds < 6)  # 6 to give a bit of leeway in case takes too long

    def test_view_name_filter(self):
        requests = [mock_suite.mock_request() for _ in range(0, 10)]
        r = random.choice(requests)
        view_name = r.view_name
        requests = models.Request.objects.filter(ViewNameFilter(view_name))
        for r in requests:
            self.assertTrue(r.view_name == view_name)

    def test_path_filter(self):
        requests = [mock_suite.mock_request() for _ in range(0, 10)]
        r = random.choice(requests)
        path = r.path
        requests = models.Request.objects.filter(PathFilter(path))
        for r in requests:
            self.assertTrue(r.path == path)

    def test_num_queries_filter(self):
        requests = [mock_suite.mock_request() for _ in range(0, 10)]
        counts = sorted(x.queries.count() for x in requests)
        c = counts[int(floor(len(counts) / 2))]
        num_queries_filter = NumQueriesFilter(c)
        query_set = models.Request.objects.all()
        query_set = num_queries_filter.contribute_to_query_set(query_set)
        filtered = query_set.filter(num_queries_filter)
        for f in filtered:
            self.assertGreaterEqual(f.queries.count(), c)

    def test_time_spent_queries_filter(self):
        requests = [mock_suite.mock_request() for _ in range(0, 10)]
        time_taken = sorted(sum(q.time_taken for q in x.queries.all()) for x in requests)
        c = time_taken[int(floor(len(time_taken) / 2))]
        time_taken_filter = TimeSpentOnQueriesFilter(c)
        query_set = models.Request.objects.all()
        query_set = time_taken_filter.contribute_to_query_set(query_set)
        filtered = query_set.filter(time_taken_filter)
        for f in filtered:
            self.assertGreaterEqual(sum(q.time_taken for q in f.queries.all()), c)

    def test_time_spent_filter(self):
        requests = [mock_suite.mock_request() for _ in range(0, 10)]
        time_taken = sorted(x.time_taken for x in requests)
        c = time_taken[int(floor(len(time_taken) / 2))]
        time_taken_filter = OverallTimeFilter(c)
        query_set = models.Request.objects.all()
        query_set = time_taken_filter.contribute_to_query_set(query_set)
        filtered = query_set.filter(time_taken_filter)
        for f in filtered:
            self.assertGreaterEqual(round(f.time_taken), round(c))

    def test_status_code_filter(self):
        requests = [mock_suite.mock_request() for _ in range(0, 50)]
        requests = sorted(requests, key=lambda x: x.response.status_code)
        by_status_code = groupby(requests, key=lambda x: x.response.status_code)
        for status_code, expected in by_status_code:
            status_code_filter = StatusCodeFilter(status_code)
            query_set = models.Request.objects.all()
            query_set = status_code_filter.contribute_to_query_set(query_set)
            filtered = query_set.filter(status_code_filter)
            self.assertEqual(len(list(expected)), filtered.count())

    def test_method_filter(self):
        requests = [mock_suite.mock_request() for _ in range(0, 50)]
        requests = sorted(requests, key=lambda x: x.method)
        by_method = groupby(requests, key=lambda x: x.method)
        for method, expected in by_method:
            method_filter = MethodFilter(method)
            query_set = models.Request.objects.all()
            query_set = method_filter.contribute_to_query_set(query_set)
            filtered = query_set.filter(method_filter)
            self.assertEqual(len(list(expected)), filtered.count())


class TestRequestAfterDateFilter(TestCase):
    def assertFilter(self, dt, f):
        requests = models.Request.objects.filter(f)
        for r in requests:
            self.assertTrue(r.start_time > dt)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.requests = [mock_suite.mock_request() for _ in range(0, 10)]

    def test_after_date_filter(self):
        r = random.choice(self.requests)
        dt = r.start_time
        f = AfterDateFilter(dt)
        self.assertFilter(dt, f)

    def test_after_date_filter_str(self):
        r = random.choice(self.requests)
        dt = r.start_time
        fmt = AfterDateFilter.fmt
        dt_str = dt.strftime(fmt)
        date_filter = AfterDateFilter
        f = date_filter(dt_str)
        new_dt = datetime.strptime(dt_str, fmt)
        new_dt = timezone.make_aware(new_dt, pytz.UTC)
        self.assertFilter(new_dt, f)


class TestRequestBeforeDateFilter(TestCase):
    def assertFilter(self, dt, f):
        requests = models.Request.objects.filter(f)
        for r in requests:
            self.assertTrue(r.start_time < dt)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.requests = [mock_suite.mock_request() for _ in range(0, 10)]

    def test_before_date_filter(self):
        r = random.choice(self.requests)
        dt = r.start_time
        f = BeforeDateFilter(dt)
        self.assertFilter(dt, f)

    def test_before_date_filter_str(self):
        r = random.choice(self.requests)
        dt = r.start_time
        fmt = BeforeDateFilter.fmt
        dt_str = dt.strftime(fmt)
        date_filter = BeforeDateFilter
        f = date_filter(dt_str)
        new_dt = datetime.strptime(dt_str, fmt)
        new_dt = timezone.make_aware(new_dt, pytz.UTC)
        self.assertFilter(new_dt, f)


class TestProfileFilters(TestCase):
    def setUp(self):
        delete_all_models(models.Profile)

    def test_name_filter(self):
        profiles = mock_suite.mock_profiles(n=10)
        p = random.choice(profiles)
        name = p.name
        requests = models.Profile.objects.filter(NameFilter(name))
        for p in requests:
            self.assertTrue(p.name == name)

    def test_function_name_filter(self):
        profiles = mock_suite.mock_profiles(n=10)
        p = random.choice(profiles)
        func_name = p.func_name
        requests = models.Profile.objects.filter(FunctionNameFilter(func_name))
        for p in requests:
            self.assertTrue(p.func_name == func_name)

    def test_num_queries_filter(self):
        profiles = mock_suite.mock_profiles(n=10)
        counts = sorted(x.queries.count() for x in profiles)
        c = counts[int(floor(len(counts) / 2))]
        num_queries_filter = NumQueriesFilter(c)
        query_set = models.Profile.objects.all()
        query_set = num_queries_filter.contribute_to_query_set(query_set)
        filtered = query_set.filter(num_queries_filter)
        for f in filtered:
            self.assertGreaterEqual(f.queries.count(), c)

    def test_time_spent_queries_filter(self):
        profiles = mock_suite.mock_profiles(n=10)
        time_taken = sorted(sum(q.time_taken for q in x.queries.all()) for x in profiles)
        c = time_taken[int(floor(len(time_taken) / 2))]
        time_taken_filter = TimeSpentOnQueriesFilter(c)
        query_set = models.Profile.objects.all()
        query_set = time_taken_filter.contribute_to_query_set(query_set)
        filtered = query_set.filter(time_taken_filter)
        for f in filtered:
            self.assertGreaterEqual(sum(q.time_taken for q in f.queries.all()), c)

    def test_time_spent_filter(self):
        profiles = [mock_suite.mock_request() for _ in range(0, 10)]
        time_taken = sorted(x.time_taken for x in profiles)
        c = time_taken[int(floor(len(time_taken) / 2))]
        time_taken_filter = OverallTimeFilter(c)
        query_set = models.Profile.objects.all()
        query_set = time_taken_filter.contribute_to_query_set(query_set)
        filtered = query_set.filter(time_taken_filter)
        for f in filtered:
            self.assertGreaterEqual(f.time_taken, c)



import base64
import json
import logging
import re
import sys
import traceback
from uuid import UUID

from django.core.exceptions import RequestDataTooBig
from django.urls import Resolver404, resolve

from silk import models
from silk.collector import DataCollector
from silk.config import SilkyConfig

Logger = logging.getLogger('silk.model_factory')

content_types_json = ['application/json',
                      'application/x-javascript',
                      'text/javascript',
                      'text/x-javascript',
                      'text/x-json']
content_type_form = ['multipart/form-data',
                     'application/x-www-form-urlencoded']
content_type_html = ['text/html']
content_type_css = ['text/css']


def _get_response_headers(response):
    """
    Django 3.2 (more specifically, commit bcc2befd0e9c1885e45b46d0b0bcdc11def8b249) broke the usage of _headers, which
    were turned into a public interface, so we need this compatibility wrapper.
    """
    try:
        return response.headers
    except AttributeError:
        return response._headers


class DefaultEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, UUID):
            return str(o)


def _parse_content_type(content_type):
    """best efforts on pulling out the content type and encoding from Content-Type header"""
    try:
        content_type = content_type.strip()
    except AttributeError:
        pass
    char_set = None
    if content_type.strip():
        splt = content_type.split(';')
        content_type = splt[0]
        try:
            raw_char_set = splt[1].strip()
            key, char_set = raw_char_set.split('=')
            if key != 'charset':
                char_set = None
        except (IndexError, ValueError):
            pass
    return content_type, char_set


class RequestModelFactory:
    """Produce Request models from Django request objects"""
    # String to replace on masking
    CLEANSED_SUBSTITUTE = '********************'

    def __init__(self, request):
        super().__init__()
        self.request = request

    def content_type(self):
        content_type = self.request.META.get('CONTENT_TYPE', '')
        return _parse_content_type(content_type)

    def encoded_headers(self):
        """
        From Django docs (https://docs.djangoproject.com/en/2.0/ref/request-response/#httprequest-objects):

        "With the exception of CONTENT_LENGTH and CONTENT_TYPE, as given above, any HTTP headers in the request are converted to
        META keys by converting all characters to uppercase, replacing any hyphens with underscores and adding an HTTP_ prefix
        to the name. So, for example, a header called X-Bender would be mapped to the META key HTTP_X_BENDER."
        """
        headers = {}
        sensitive_headers = {'AUTHORIZATION'}

        for k, v in self.request.META.items():
            if k.startswith('HTTP') or k in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
                splt = k.split('_')
                if splt[0] == 'HTTP':
                    splt = splt[1:]
                k = '-'.join(splt)

                if k in sensitive_headers:
                    v = RequestModelFactory.CLEANSED_SUBSTITUTE

                headers[k] = v
        if SilkyConfig().SILKY_HIDE_COOKIES:
            try:
                del headers['COOKIE']
            except KeyError:
                pass

        return json.dumps(headers, cls=DefaultEncoder, ensure_ascii=SilkyConfig().SILKY_JSON_ENSURE_ASCII)

    def _mask_credentials(self, body):
        """
        Mask credentials of potentially sensitive info before saving to db.
        """
        sensitive_keys = SilkyConfig().SILKY_SENSITIVE_KEYS
        key_string = '|'.join(sensitive_keys)

        def replace_pattern_values(obj):
            pattern = re.compile(key_string, re.I)
            if isinstance(obj, dict):
                for key in obj.keys():
                    if pattern.search(key):
                        obj[key] = RequestModelFactory.CLEANSED_SUBSTITUTE
                    else:
                        obj[key] = replace_pattern_values(obj[key])
            elif isinstance(obj, list):
                for index, item in enumerate(obj):
                    obj[index] = replace_pattern_values(item)
            else:
                if pattern.search(str(obj)):
                    return RequestModelFactory.CLEANSED_SUBSTITUTE
            return obj

        try:
            json_body = json.loads(body)
        except Exception as e:
            pattern = re.compile(fr'(({key_string})[^=]*)=(.*?)(&|$)', re.M | re.I)
            try:
                body = re.sub(pattern, f'\\1={RequestModelFactory.CLEANSED_SUBSTITUTE}\\4', body)
            except Exception:
                Logger.debug(f'{str(e)}')
        else:
            body = json.dumps(replace_pattern_values(json_body), ensure_ascii=SilkyConfig().SILKY_JSON_ENSURE_ASCII)

        return body

    def _body(self, raw_body, content_type):
        """
        Encode body as JSON if possible so can be used as a dictionary in generation
        of curl/django test client code
        """
        body = ''
        if content_type in content_type_form:
            body = self.request.POST
            body = json.dumps(dict(body), sort_keys=True, indent=4
                              , ensure_ascii=SilkyConfig().SILKY_JSON_ENSURE_ASCII)
        elif content_type in content_types_json:
            try:
                body = json.dumps(json.loads(raw_body), sort_keys=True, indent=4
                                  , ensure_ascii=SilkyConfig().SILKY_JSON_ENSURE_ASCII)
            except Exception:
                body = raw_body
        return body

    def body(self):
        content_type, char_set = self.content_type()
        try:
            raw_body = self.request.body
        except RequestDataTooBig:
            raw_body = b"Raw body exceeds DATA_UPLOAD_MAX_MEMORY_SIZE, Silk is not showing file uploads."
            body = self.request.POST.copy()
            for k, v in self.request.FILES.items():
                body.appendlist(k, v)
            return body, raw_body
        if char_set:
            try:
                raw_body = raw_body.decode(char_set)
            except AttributeError:
                pass
            except LookupError:  # If no encoding exists, default to UTF-8
                try:
                    raw_body = raw_body.decode('UTF-8')
                except AttributeError:
                    pass
                except UnicodeDecodeError:
                    raw_body = ''
            except Exception as e:
                Logger.error(
                    'Unable to decode request body using char_set %s due to error: %s. Will ignore. Stacktrace:'
                    % (char_set, e)
                )
                traceback.print_exc()
        else:
            # Default to an attempt at UTF-8 decoding.
            try:
                raw_body = raw_body.decode('UTF-8')
            except AttributeError:
                pass
            except UnicodeDecodeError:
                raw_body = ''
        max_size = SilkyConfig().SILKY_MAX_REQUEST_BODY_SIZE
        body = ''
        if raw_body:
            if max_size > -1:
                Logger.debug('A max request size is set so checking size')
                size = sys.getsizeof(raw_body, default=None)
                request_identifier = self.request.path
                if not size:
                    Logger.error(
                        'No way in which to get size of request body for %s, will ignore it',
                        request_identifier
                    )
                elif size <= max_size:
                    Logger.debug(
                        'Request %s has body of size %d which is less than %d so will save the body'
                        % (request_identifier, size, max_size)
                    )
                    body = self._body(raw_body, content_type)
                else:
                    Logger.debug(
                        'Request %s has body of size %d which is greater than %d, therefore ignoring'
                        % (request_identifier, size, max_size)
                    )
                    raw_body = None
            else:
                Logger.debug('No maximum request body size is set, continuing.')
                body = self._body(raw_body, content_type)
        body = self._mask_credentials(body)
        raw_body = self._mask_credentials(raw_body)
        return body, raw_body

    def query_params(self):
        query_params = self.request.GET
        encoded_query_params = ''
        if query_params:
            query_params_dict = dict(zip(query_params.keys(), query_params.values()))
            encoded_query_params = json.dumps(query_params_dict, ensure_ascii=SilkyConfig().SILKY_JSON_ENSURE_ASCII)
        return encoded_query_params

    def view_name(self):
        try:
            resolved = resolve(self.request.path_info)
        except Resolver404:
            return None

        return resolved.view_name

    def construct_request_model(self):
        body, raw_body = self.body()
        query_params = self.query_params()
        path = self.request.path
        view_name = self.view_name()

        request_model = models.Request.objects.create(
            path=path,
            encoded_headers=self.encoded_headers(),
            method=self.request.method,
            query_params=query_params,
            view_name=view_name,
            body=body)
        # Text fields are encoded as UTF-8 in Django and hence will try to coerce
        # anything to we pass to UTF-8. Some stuff like binary will fail.
        try:
            request_model.raw_body = raw_body
        except UnicodeDecodeError:
            Logger.debug('NYI: Binary request bodies')  # TODO
        Logger.debug('Created new request model with pk %s' % request_model.pk)
        return request_model


class ResponseModelFactory:
    """given a response object, craft the silk response model"""

    def __init__(self, response):
        super().__init__()
        self.response = response
        self.request = DataCollector().request

    def body(self):
        body = ''
        content_type, char_set = _parse_content_type(self.response.get('Content-Type', ''))
        content = getattr(self.response, 'content', '')
        if content:
            max_body_size = SilkyConfig().SILKY_MAX_RESPONSE_BODY_SIZE
            if max_body_size > -1:
                Logger.debug('Max size of response body defined so checking')
                size = sys.getsizeof(content, None)
                if not size:
                    Logger.error('Could not get size of response body. Ignoring')
                    content = ''
                else:
                    if size > max_body_size:
                        content = ''
                        Logger.debug(
                            'Size of %d for %s is bigger than %d so ignoring response body'
                            % (size, self.request.path, max_body_size)
                        )
                    else:
                        Logger.debug(
                            'Size of %d for %s is less than %d so saving response body'
                            % (size, self.request.path, max_body_size)
                        )
            if content and content_type in content_types_json:
                # TODO: Perhaps theres a way to format the JSON without parsing it?
                if not isinstance(content, str):
                    # byte string is not compatible with json.loads(...)
                    # and json.dumps(...) in python3
                    content = content.decode()
                try:
                    body = json.dumps(json.loads(content), sort_keys=True, indent=4
                                      , ensure_ascii=SilkyConfig().SILKY_JSON_ENSURE_ASCII)
                except (TypeError, ValueError):
                    Logger.warn(
                        'Response to request with pk %s has content type %s but was unable to parse it'
                        % (self.request.pk, content_type)
                    )
        return body, content

    def construct_response_model(self):
        assert self.request, 'Cant construct a response model if there is no request model'
        Logger.debug(
            'Creating response model for request model with pk %s'
            % self.request.pk
        )
        b, content = self.body()
        raw_headers = _get_response_headers(self.response)
        headers = {}
        for k, v in raw_headers.items():
            try:
                header, val = v
            except ValueError:
                header, val = k, v
            finally:
                headers[header] = val
        silky_response = models.Response.objects.create(
            request_id=self.request.id,
            status_code=self.response.status_code,
            encoded_headers=json.dumps(headers, ensure_ascii=SilkyConfig().SILKY_JSON_ENSURE_ASCII),
            body=b
        )

        try:
            raw_body = base64.b64encode(content)
        except TypeError:
            raw_body = base64.b64encode(content.encode('utf-8'))
        silky_response.raw_body = raw_body.decode('ascii')
        return silky_response



import logging
import random

from django.db import DatabaseError, transaction
from django.db.models.sql.compiler import SQLCompiler
from django.urls import NoReverseMatch, reverse
from django.utils import timezone

from silk.collector import DataCollector
from silk.config import SilkyConfig
from silk.model_factory import RequestModelFactory, ResponseModelFactory
from silk.profiling import dynamic
from silk.profiling.profiler import silk_meta_profiler
from silk.sql import execute_sql

Logger = logging.getLogger('silk.middleware')


def silky_reverse(name, *args, **kwargs):
    try:
        r = reverse('silk:%s' % name, *args, **kwargs)
    except NoReverseMatch:
        # In case user forgets to set namespace, but also fixes Django 1.5 tests on Travis
        # Hopefully if user has forgotten to add namespace there are no clashes with their own
        # view names but I don't think there is really anything can do about this.
        r = reverse(name, *args, **kwargs)
    return r


def get_fpath():
    return silky_reverse('summary')


config = SilkyConfig()


def _should_intercept(request):
    """we want to avoid recording any requests/sql queries etc that belong to Silky"""
    # Check custom intercept logic.
    if config.SILKY_INTERCEPT_FUNC:
        if not config.SILKY_INTERCEPT_FUNC(request):
            return False
    # don't trap every request
    elif config.SILKY_INTERCEPT_PERCENT < 100:
        if random.random() > config.SILKY_INTERCEPT_PERCENT / 100.0:
            return False

    try:
        silky = request.path.startswith(get_fpath())
    except NoReverseMatch:
        silky = False

    ignored = request.path in config.SILKY_IGNORE_PATHS
    return not (silky or ignored)


class TestMiddleware:
    def process_response(self, request, response):
        return response

    def process_request(self, request):
        return


class SilkyMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        self.process_request(request)

        response = self.get_response(request)

        response = self.process_response(request, response)

        return response

    def _apply_dynamic_mappings(self):
        dynamic_profile_configs = config.SILKY_DYNAMIC_PROFILING
        for conf in dynamic_profile_configs:
            module = conf.get('module')
            function = conf.get('function')
            start_line = conf.get('start_line')
            end_line = conf.get('end_line')
            name = conf.get('name')
            if module and function:
                if start_line and end_line:  # Dynamic context manager
                    dynamic.inject_context_manager_func(module=module,
                                                        func=function,
                                                        start_line=start_line,
                                                        end_line=end_line,
                                                        name=name)
                else:  # Dynamic decorator
                    dynamic.profile_function_or_method(module=module,
                                                       func=function,
                                                       name=name)
            else:
                raise KeyError('Invalid dynamic mapping %s' % conf)

    @silk_meta_profiler()
    def process_request(self, request):
        DataCollector().clear()

        if not _should_intercept(request):
            return

        Logger.debug('process_request')
        request.silk_is_intercepted = True
        self._apply_dynamic_mappings()
        if not hasattr(SQLCompiler, '_execute_sql'):
            SQLCompiler._execute_sql = SQLCompiler.execute_sql
            SQLCompiler.execute_sql = execute_sql

        silky_config = SilkyConfig()

        should_profile = silky_config.SILKY_PYTHON_PROFILER
        if silky_config.SILKY_PYTHON_PROFILER_FUNC:
            should_profile = silky_config.SILKY_PYTHON_PROFILER_FUNC(request)

        request_model = RequestModelFactory(request).construct_request_model()
        DataCollector().configure(request_model, should_profile=should_profile)

    @transaction.atomic()
    def _process_response(self, request, response):
        Logger.debug('Process response')
        with silk_meta_profiler():
            collector = DataCollector()
            collector.stop_python_profiler()
            silk_request = collector.request
            if silk_request:
                silk_response = ResponseModelFactory(response).construct_response_model()
                silk_response.save()
                silk_request.end_time = timezone.now()
                collector.finalise()
            else:
                Logger.error(
                    'No request model was available when processing response. '
                    'Did something go wrong in process_request/process_view?'
                    '\n' + str(request) + '\n\n' + str(response)
                )
        # Need to save the data outside the silk_meta_profiler
        # Otherwise the  meta time collected in the context manager
        # is not taken in account
        if silk_request:
            silk_request.save()
        Logger.debug('Process response done.')

    def process_response(self, request, response):
        if getattr(request, 'silk_is_intercepted', False):
            while True:
                try:
                    self._process_response(request, response)
                except (AttributeError, DatabaseError):
                    Logger.debug('Retrying _process_response')
                    self._process_response(request, response)
                finally:
                    break
        return response



import logging
import traceback

from django.core.exceptions import EmptyResultSet
from django.utils import timezone

from silk.collector import DataCollector
from silk.config import SilkyConfig

Logger = logging.getLogger('silk.sql')


def _should_wrap(sql_query):
    if not DataCollector().request:
        return False

    for ignore_str in SilkyConfig().SILKY_IGNORE_QUERIES:
        if ignore_str in sql_query:
            return False
    return True


def _unpack_explanation(result):
    for row in result:
        if not isinstance(row, str):
            yield ' '.join(str(c) for c in row)
        else:
            yield row


def _explain_query(connection, q, params):
    if connection.features.supports_explaining_query_execution:
        if SilkyConfig().SILKY_ANALYZE_QUERIES:
            # Work around some DB engines not supporting analyze option
            try:
                prefix = connection.ops.explain_query_prefix(
                    analyze=True, **(SilkyConfig().SILKY_EXPLAIN_FLAGS or {})
                )
            except ValueError as error:
                error_str = str(error)
                if error_str.startswith("Unknown options:"):
                    Logger.warning(
                        "Database does not support analyzing queries with provided params. %s."
                        "SILKY_ANALYZE_QUERIES option will be ignored",
                        error_str
                    )
                    prefix = connection.ops.explain_query_prefix()
                else:
                    raise error
        else:
            prefix = connection.ops.explain_query_prefix()

        # currently we cannot use explain() method
        # for queries other than `select`
        prefixed_query = f"{prefix} {q}"
        with connection.cursor() as cur:
            cur.execute(prefixed_query, params)
            result = _unpack_explanation(cur.fetchall())
            return '\n'.join(result)
    return None


def execute_sql(self, *args, **kwargs):
    """wrapper around real execute_sql in order to extract information"""

    try:
        q, params = self.as_sql()
        if not q:
            raise EmptyResultSet
    except EmptyResultSet:
        try:
            result_type = args[0]
        except IndexError:
            result_type = kwargs.get('result_type', 'multi')
        if result_type == 'multi':
            return iter([])
        else:
            return
    tb = ''.join(reversed(traceback.format_stack()))
    sql_query = q % params
    if _should_wrap(sql_query):
        query_dict = {
            'query': sql_query,
            'start_time': timezone.now(),
            'traceback': tb
        }
        try:
            return self._execute_sql(*args, **kwargs)
        finally:
            query_dict['end_time'] = timezone.now()
            request = DataCollector().request
            if request:
                query_dict['request'] = request
            if self.query.model.__module__ != 'silk.models':
                query_dict['analysis'] = _explain_query(self.connection, q, params)
                DataCollector().register_query(query_dict)
            else:
                DataCollector().register_silk_query(query_dict)
    return self._execute_sql(*args, **kwargs)



from unittest.mock import Mock

from silk.models import Request


def mock_data_collector():
    mock = Mock()
    mock.queries = []
    mock.local = Mock()
    r = Request()
    mock.local.request = r
    mock.request = r
    return mock


def delete_all_models(model_class):
    """
    A sqlite3-safe deletion function to avoid "django.db.utils.OperationalError: too many SQL variables"
    :param model_class:
    :return:
    """
    while model_class.objects.count():
        ids = model_class.objects.values_list('pk', flat=True)[:80]
        model_class.objects.filter(pk__in=ids).delete()



from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.models import Request


class CProfileView(View):

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, *_, **kwargs):
        request_id = kwargs['request_id']
        silk_request = Request.objects.get(pk=request_id)
        context = {
            'silk_request': silk_request,
            'request': request}

        return render(request, 'silk/cprofile.html', context)



import json
from unittest.mock import Mock

from django.test import TestCase

from silk.model_factory import RequestModelFactory, ResponseModelFactory

DJANGO_META_CONTENT_TYPE = 'CONTENT_TYPE'
HTTP_CONTENT_TYPE = 'Content-Type'


class TestEncodingForRequests(TestCase):
    """
    Check that the RequestModelFactory deals with encodings correctly via charset
    """

    def test_utf_plain(self):
        mock_request = Mock()
        mock_request.META = {DJANGO_META_CONTENT_TYPE: 'text/plain; charset=UTF-8'}
        mock_request.body = '语'
        mock_request.get = mock_request.META.get
        factory = RequestModelFactory(mock_request)
        body, raw_body = factory.body()
        self.assertFalse(body)
        self.assertEqual(raw_body, mock_request.body)

    def test_plain(self):
        mock_request = Mock()
        mock_request.META = {DJANGO_META_CONTENT_TYPE: 'text/plain'}
        mock_request.body = 'sdfsdf'
        mock_request.get = mock_request.META.get
        factory = RequestModelFactory(mock_request)
        body, raw_body = factory.body()
        self.assertFalse(body)
        self.assertEqual(raw_body, mock_request.body)

    def test_utf_json_not_encoded(self):
        mock_request = Mock()
        mock_request.META = {DJANGO_META_CONTENT_TYPE: 'application/json; charset=UTF-8'}
        d = {'x': '语'}
        mock_request.body = json.dumps(d)
        mock_request.get = mock_request.META.get
        factory = RequestModelFactory(mock_request)
        body, raw_body = factory.body()
        self.assertDictEqual(json.loads(body), d)
        self.assertEqual(raw_body, mock_request.body)

    def test_utf_json_encoded(self):
        mock_request = Mock()
        mock_request.META = {DJANGO_META_CONTENT_TYPE: 'application/json; charset=UTF-8'}
        d = {'x': '语'}
        mock_request.body = json.dumps(d).encode('UTF-8')
        mock_request.get = mock_request.META.get
        factory = RequestModelFactory(mock_request)
        body, raw_body = factory.body()
        self.assertDictEqual(json.loads(body), d)
        self.assertEqual(raw_body, mock_request.body.decode('UTF-8'))

    def test_utf_json_encoded_no_charset(self):
        """default to UTF-8"""
        mock_request = Mock()
        mock_request.META = {DJANGO_META_CONTENT_TYPE: 'application/json'}
        d = {'x': '语'}
        mock_request.body = json.dumps(d).encode('UTF-8')
        mock_request.get = mock_request.META.get
        factory = RequestModelFactory(mock_request)
        body, raw_body = factory.body()
        self.assertDictEqual(json.loads(body), d)
        self.assertEqual(raw_body, mock_request.body.decode('UTF-8'))

    def test_invalid_encoding_json(self):
        mock_request = Mock()
        mock_request.META = {DJANGO_META_CONTENT_TYPE: 'application/json; charset=asdas-8'}
        d = {'x': '语'}
        mock_request.body = json.dumps(d).encode('UTF-8')
        mock_request.get = mock_request.META.get
        factory = RequestModelFactory(mock_request)
        body, raw_body = factory.body()
        self.assertDictEqual(json.loads(body), d)
        self.assertEqual(raw_body, raw_body)


class TestEncodingForResponse(TestCase):
    """
    Check that the ResponseModelFactory deals with encodings correctly via charset
    """

    def test_utf_plain(self):
        mock = Mock()
        mock.headers = {HTTP_CONTENT_TYPE: 'text/plain; charset=UTF-8'}
        mock.content = '语'
        mock.get = mock.headers.get
        factory = ResponseModelFactory(mock)
        body, content = factory.body()
        self.assertFalse(body)
        self.assertEqual(content, mock.content)

    def test_plain(self):
        mock = Mock()
        mock.headers = {HTTP_CONTENT_TYPE: 'text/plain'}
        mock.content = 'sdfsdf'
        mock.get = mock.headers.get
        factory = ResponseModelFactory(mock)
        body, content = factory.body()
        self.assertFalse(body)
        self.assertEqual(content, mock.content)

    def test_utf_json_not_encoded(self):
        mock = Mock()
        mock.headers = {HTTP_CONTENT_TYPE: 'application/json; charset=UTF-8'}
        d = {'x': '语'}
        mock.content = json.dumps(d)
        mock.get = mock.headers.get
        factory = ResponseModelFactory(mock)
        body, content = factory.body()
        self.assertDictEqual(json.loads(body), d)
        self.assertEqual(content, mock.content)

    def test_utf_json_encoded(self):
        mock = Mock()
        mock.headers = {HTTP_CONTENT_TYPE: 'application/json; charset=UTF-8'}
        d = {'x': '语'}
        mock.content = json.dumps(d)
        mock.get = mock.headers.get
        factory = ResponseModelFactory(mock)
        body, content = factory.body()
        self.assertDictEqual(json.loads(body), d)
        self.assertEqual(content, mock.content)

    def test_utf_json_encoded_no_charset(self):
        """default to UTF-8"""
        mock = Mock()
        mock.headers = {HTTP_CONTENT_TYPE: 'application/json'}
        d = {'x': '语'}
        mock.content = json.dumps(d)
        mock.get = mock.headers.get
        factory = ResponseModelFactory(mock)
        body, content = factory.body()
        self.assertDictEqual(json.loads(body), d)
        self.assertEqual(content, mock.content)

    def test_invalid_encoding_json(self):
        mock = Mock()
        mock.headers = {HTTP_CONTENT_TYPE: 'application/json; charset=asdas-8'}
        d = {'x': '语'}
        mock.content = json.dumps(d)
        mock.get = mock.headers.get
        factory = ResponseModelFactory(mock)
        body, content = factory.body()
        self.assertDictEqual(json.loads(body), d)
        self.assertEqual(mock.content, content)



# Generated by Django 3.2 on 2021-04-12 22:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('example_app', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='blind',
            name='photo',
            field=models.ImageField(upload_to='products'),
        ),
    ]



from unittest.mock import Mock

from django.test import TestCase
from django.urls import reverse

from silk.collector import DataCollector
from silk.config import SilkyConfig
from silk.model_factory import RequestModelFactory, ResponseModelFactory
from silk.models import Request


class TestMaxBodySizeRequest(TestCase):

    def test_no_max_request(self):
        SilkyConfig().SILKY_MAX_REQUEST_BODY_SIZE = -1
        mock_request = Mock()
        mock_request.META = {'CONTENT_TYPE': 'text/plain'}
        mock_request.GET = {}
        mock_request.path = reverse('silk:requests')
        mock_request.method = 'get'
        mock_request.body = b'a' * 1000  # 1000 bytes?
        request_model = RequestModelFactory(mock_request).construct_request_model()
        self.assertTrue(request_model.raw_body)

    def test_max_request(self):
        SilkyConfig().SILKY_MAX_REQUEST_BODY_SIZE = 10  # 10kb
        mock_request = Mock()
        mock_request.META = {'CONTENT_TYPE': 'text/plain'}
        mock_request.GET = {}
        mock_request.method = 'get'
        mock_request.body = b'a' * 1024 * 100  # 100kb
        mock_request.path = reverse('silk:requests')
        request_model = RequestModelFactory(mock_request).construct_request_model()
        self.assertFalse(request_model.raw_body)


class TestMaxBodySizeResponse(TestCase):

    def setUp(self):
        DataCollector().request = Request.objects.create()

    def test_no_max_response(self):
        SilkyConfig().SILKY_MAX_RESPONSE_BODY_SIZE = -1
        mock_response = Mock()
        headers = {'CONTENT_TYPE': 'text/plain'}
        mock_response.get = headers.get
        mock_response.headers = headers
        mock_response.content = b'a' * 1000  # 1000 bytes?
        mock_response.status_code = 200
        response_model = ResponseModelFactory(mock_response).construct_response_model()
        self.assertTrue(response_model.raw_body)

    def test_max_response(self):
        SilkyConfig().SILKY_MAX_RESPONSE_BODY_SIZE = 10  # 10kb
        mock_response = Mock()
        headers = {'CONTENT_TYPE': 'text/plain'}
        mock_response.get = headers.get
        mock_response.headers = headers
        mock_response.content = b'a' * 1024 * 100  # 100kb
        mock_response.status_code = 200
        response_model = ResponseModelFactory(mock_response).construct_response_model()
        self.assertFalse(response_model.raw_body)



from unittest.mock import Mock

from django.test import TestCase

from silk.model_factory import RequestModelFactory


class TestLongRequestUrl(TestCase):

    def test_no_long_url(self):
        url = '1234567890' * 19  # 190-character URL
        mock_request = Mock()
        mock_request.META = {'CONTENT_TYPE': 'text/plain'}
        mock_request.GET = {}
        mock_request.path = url
        mock_request.method = 'get'
        request_model = RequestModelFactory(mock_request).construct_request_model()
        self.assertEqual(request_model.path, url)

    def test_long_url(self):
        url = '1234567890' * 200  # 2000-character URL
        mock_request = Mock()
        mock_request.META = {'CONTENT_TYPE': 'text/plain'}
        mock_request.GET = {}
        mock_request.method = 'get'
        mock_request.path = url
        request_model = RequestModelFactory(mock_request).construct_request_model()
        self.assertEqual(request_model.path, f'{url[:94]}...{url[1907:]}')
        self.assertEqual(len(request_model.path), 190)



from django import template
from django.urls import reverse

register = template.Library()


@register.simple_tag
def navactive(request, urls, *args, **kwargs):
    path = request.path
    urls = [reverse(url, args=args) for url in urls.split()]
    if path in urls:
        cls = kwargs.get('class', None)
        if not cls:
            cls = "menu-item-selected"
        return cls
    return ""



# Generated by Django 2.2.17 on 2020-11-26 13:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('silk', '0007_sqlquery_identifier'),
    ]

    operations = [
        migrations.AddField(
            model_name='sqlquery',
            name='analysis',
            field=models.TextField(blank=True, null=True),
        ),
    ]



import re

_pattern = re.compile(' +')


def parse_profile(output):
    """
    Parse the output of cProfile to a list of tuples.
    """
    if isinstance(output, str):
        output = output.split('\n')
    for i, line in enumerate(output):
        # ignore n function calls, total time and ordered by and empty lines
        line = line.strip()
        if i > 3 and line:
            columns = _pattern.split(line)[0:]
            function = ' '.join(columns[5:])
            columns = columns[:5] + [function]
            yield columns



import factory
import factory.fuzzy
from example_app.models import Blind

from silk.models import Request, Response, SQLQuery

HTTP_METHODS = ['GET', 'POST', 'PUT', 'PATCH', 'HEAD', 'OPTIONS']
STATUS_CODES = [200, 201, 300, 301, 302, 401, 403, 404]


class SQLQueryFactory(factory.django.DjangoModelFactory):

    query = factory.Sequence(lambda num: 'SELECT foo FROM bar WHERE foo=%s' % num)
    traceback = factory.Sequence(lambda num: 'Traceback #%s' % num)

    class Meta:
        model = SQLQuery


class RequestMinFactory(factory.django.DjangoModelFactory):

    path = factory.Faker('uri_path')
    method = factory.fuzzy.FuzzyChoice(HTTP_METHODS)

    class Meta:
        model = Request


class ResponseFactory(factory.django.DjangoModelFactory):
    request = factory.SubFactory(RequestMinFactory)
    status_code = factory.fuzzy.FuzzyChoice(STATUS_CODES)

    class Meta:
        model = Response


class BlindFactory(factory.django.DjangoModelFactory):
    name = factory.Faker('pystr', min_chars=5, max_chars=10)
    child_safe = factory.Faker('pybool')
    photo = factory.django.ImageField()

    class Meta:
        model = Blind



from django.db.models import Count, Sum
from django.shortcuts import render
from django.template.context_processors import csrf
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.models import Profile, Request
from silk.request_filters import BaseFilter, filters_from_request


class ProfilingView(View):
    show = [5, 10, 25, 100, 250]
    default_show = 25
    order_by = ['Recent',
                'Name',
                'Function Name',
                'Num. Queries',
                'Time',
                'Time on queries']
    defualt_order_by = 'Recent'
    session_key_profile_filters = 'session_key_profile_filters'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _get_distinct_values(self, field, silk_request):
        if silk_request:
            query_set = Profile.objects.filter(request=silk_request)
        else:
            query_set = Profile.objects.all()
        function_names = [x[field] for x in query_set.values(field).distinct()]
        # Ensure top, default option is ''
        try:
            function_names.remove('')
        except ValueError:
            pass
        return [''] + function_names

    def _get_function_names(self, silk_request=None):
        return self._get_distinct_values('func_name', silk_request)

    def _get_names(self, silk_request=None):
        return self._get_distinct_values('name', silk_request)

    def _get_objects(self, show=None, order_by=None, name=None, func_name=None, silk_request=None, filters=None):
        if not filters:
            filters = []
        if not show:
            show = self.default_show
        manager = Profile.objects
        if silk_request:
            query_set = manager.filter(request=silk_request)
        else:
            query_set = manager.all()
        if not order_by:
            order_by = self.defualt_order_by
        if order_by == 'Recent':
            query_set = query_set.order_by('-start_time')
        elif order_by == 'Name':
            query_set = query_set.order_by('-name')
        elif order_by == 'Function Name':
            query_set = query_set.order_by('-func_name')
        elif order_by == 'Num. Queries':
            query_set = query_set.annotate(num_queries=Count('queries')).order_by('-num_queries')
        elif order_by == 'Time':
            query_set = query_set.order_by('-time_taken')
        elif order_by == 'Time on queries':
            query_set = query_set.annotate(db_time=Sum('queries__time_taken')).order_by('-db_time')
        elif order_by:
            raise RuntimeError('Unknown order_by: "%s"' % order_by)
        if func_name:
            query_set = query_set.filter(func_name=func_name)
        if name:
            query_set = query_set.filter(name=name)
        for f in filters:
            query_set = f.contribute_to_query_set(query_set)
            query_set = query_set.filter(f)
        return list(query_set[:show])

    def _create_context(self, request, *args, **kwargs):
        request_id = kwargs.get('request_id')
        if request_id:
            silk_request = Request.objects.get(pk=request_id)
        else:
            silk_request = None
        show = request.GET.get('show', self.default_show)
        order_by = request.GET.get('order_by', self.defualt_order_by)
        if show:
            show = int(show)
        func_name = request.GET.get('func_name', None)
        name = request.GET.get('name', None)
        filters = request.session.get(self.session_key_profile_filters, {})
        context = {
            'show': show,
            'order_by': order_by,
            'request': request,
            'func_name': func_name,
            'options_show': self.show,
            'options_order_by': self.order_by,
            'options_func_names': self._get_function_names(silk_request),
            'options_names': self._get_names(silk_request),
            'filters': filters
        }
        context.update(csrf(request))
        if silk_request:
            context['silk_request'] = silk_request
        if func_name:
            context['func_name'] = func_name
        if name:
            context['name'] = name
        objs = self._get_objects(show=show,
                                 order_by=order_by,
                                 func_name=func_name,
                                 silk_request=silk_request,
                                 name=name,
                                 filters=[BaseFilter.from_dict(x) for _, x in filters.items()])
        context['results'] = objs
        return context

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, *args, **kwargs):
        return render(request, 'silk/profiling.html', self._create_context(request, *args, **kwargs))

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def post(self, request):
        filters = filters_from_request(request)
        request.session[self.session_key_profile_filters] = {ident: f.as_dict() for ident, f in filters.items()}
        return render(request, 'silk/profiling.html', self._create_context(request))



import inspect
import logging
import time
import traceback
from functools import wraps

from django.conf import settings
from django.utils import timezone

from silk.collector import DataCollector
from silk.config import SilkyConfig
from silk.models import _time_taken

logger = logging.getLogger('silk.profiling.profiler')


# noinspection PyPep8Naming
class silk_meta_profiler:
    """Used in the profiling of Silk itself."""

    def __init__(self):
        super().__init__()
        self.start_time = None

    @property
    def _should_meta_profile(self):
        return SilkyConfig().SILKY_META

    def __enter__(self):
        if self._should_meta_profile:
            self.start_time = timezone.now()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._should_meta_profile:
            end_time = timezone.now()
            exception_raised = exc_type is not None
            if exception_raised:
                logger.error('Exception when performing meta profiling, dumping trace below')
                traceback.print_exception(exc_type, exc_val, exc_tb)
            request = getattr(DataCollector().local, 'request', None)
            if request:
                curr = request.meta_time or 0
                request.meta_time = curr + _time_taken(self.start_time, end_time)

    def __call__(self, target):
        if self._should_meta_profile:
            def wrapped_target(*args, **kwargs):
                request = DataCollector().request
                if request:
                    start_time = timezone.now()
                    result = target(*args, **kwargs)
                    end_time = timezone.now()
                    curr = request.meta_time or 0
                    request.meta_time = curr + _time_taken(start_time, end_time)
                else:
                    result = target(*args, **kwargs)
                return result

            return wrapped_target
        return target


# noinspection PyPep8Naming
class silk_profile:
    def __init__(self, name=None, _dynamic=False):
        super().__init__()
        self.name = name
        self.profile = None
        self._queries_before = None
        self._queries_after = None
        self._dynamic = _dynamic

    def _query_identifiers_from_collector(self):
        return [x for x in DataCollector().queries]

    def _start_queries(self):
        """record queries that have been executed before profiling began"""
        self._queries_before = self._query_identifiers_from_collector()

    def _end_queries(self):
        """record queries that have been executed after profiling has finished"""
        self._queries_after = self._query_identifiers_from_collector()

    def __enter__(self):
        if self._silk_installed() and self._should_profile():
            with silk_meta_profiler():
                self._start_queries()
                if not self.name:
                    raise ValueError('silk_profile used as a context manager must have a name')
                frame = inspect.currentframe()
                frames = inspect.getouterframes(frame)
                outer_frame = frames[1]
                path = outer_frame[1]
                line_num = outer_frame[2]
                request = DataCollector().request
                self.profile = {
                    'name': self.name,
                    'file_path': path,
                    'line_num': line_num,
                    'dynamic': self._dynamic,
                    'request': request,
                    'start_time': timezone.now(),
                }
        else:
            logger.warning('Cannot execute silk_profile as silk is not installed correctly.')

    def _finalise_queries(self):
        collector = DataCollector()
        self._end_queries()
        assert self.profile, 'no profile was created'
        diff = set(self._queries_after).difference(set(self._queries_before))
        self.profile['queries'] = diff
        collector.register_profile(self.profile)

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._silk_installed() and self._should_profile():
            with silk_meta_profiler():
                exception_raised = exc_type is not None
                self.profile['exception_raised'] = exception_raised
                self.profile['end_time'] = timezone.now()
                self._finalise_queries()

    def _silk_installed(self):
        app_installed = 'silk' in settings.INSTALLED_APPS
        middlewares = getattr(settings, 'MIDDLEWARE', [])
        if not middlewares:
            middlewares = []
        middleware_installed = SilkyConfig().SILKY_MIDDLEWARE_CLASS in middlewares
        return app_installed and middleware_installed

    def _should_profile(self):
        return DataCollector().request is not None

    def __call__(self, target):
        if self._silk_installed():
            def decorator(view_func):
                @wraps(view_func)
                def wrapped_target(*args, **kwargs):
                    with silk_meta_profiler():
                        try:
                            func_code = target.__code__
                        except AttributeError:
                            raise NotImplementedError('Profile not implemented to decorate type %s' % target.__class__.__name__)
                        line_num = func_code.co_firstlineno
                        file_path = func_code.co_filename
                        func_name = target.__name__
                        if not self.name:
                            self.name = func_name
                        self.profile = {
                            'func_name': func_name,
                            'name': self.name,
                            'file_path': file_path,
                            'line_num': line_num,
                            'dynamic': self._dynamic,
                            'start_time': timezone.now(),
                            'request': DataCollector().request
                        }
                        self._start_queries()
                    try:
                        result = target(*args, **kwargs)
                    except Exception:
                        self.profile['exception_raised'] = True
                        raise
                    finally:
                        with silk_meta_profiler():
                            self.profile['end_time'] = timezone.now()
                            self._finalise_queries()
                    return result

                return wrapped_target

            return decorator(target)
        else:
            logger.warning('Cannot execute silk_profile as silk is not installed correctly.')
            return target

    def distinct_queries(self):
        return [x for x in self._queries_after if x not in self._queries_before]


@silk_profile()
def blah():
    time.sleep(1)


if __name__ == '__main__':
    blah()



from unittest.mock import patch

from django.test import TestCase, override_settings
from django.urls import reverse

from silk.config import SilkyConfig
from silk.middleware import SilkyMiddleware, _should_intercept
from silk.models import Request

from .util import mock_data_collector


def fake_get_response():
    def fake_response():
        return 'hello world'
    return fake_response


class TestApplyDynamicMappings(TestCase):
    def test_dynamic_decorator(self):
        middleware = SilkyMiddleware(fake_get_response)
        SilkyConfig().SILKY_DYNAMIC_PROFILING = [
            {
                'module': 'tests.data.dynamic',
                'function': 'foo'
            }
        ]
        middleware._apply_dynamic_mappings()
        from .data.dynamic import foo

        mock = mock_data_collector()
        with patch('silk.profiling.profiler.DataCollector', return_value=mock) as mock_DataCollector:
            foo()  # Should be wrapped in a decorator
            self.assertTrue(mock_DataCollector.return_value.register_profile.call_count)

    def test_dynamic_context_manager(self):
        middleware = SilkyMiddleware(fake_get_response)
        SilkyConfig().SILKY_DYNAMIC_PROFILING = [
            {
                'module': 'tests.data.dynamic',
                'function': 'foo',
                'start_line': 1,
                'end_line': 2,
            }
        ]
        middleware._apply_dynamic_mappings()
        from .data.dynamic import foo

        mock = mock_data_collector()
        with patch('silk.profiling.profiler.DataCollector', return_value=mock) as mock_DataCollector:
            foo()
            self.assertTrue(mock_DataCollector.return_value.register_profile.call_count)

    def test_invalid_dynamic_context_manager(self):
        middleware = SilkyMiddleware(fake_get_response)
        SilkyConfig().SILKY_DYNAMIC_PROFILING = [
            {
                'module': 'tests.data.dynamic',
                'function': 'foo2',
                'start_line': 1,
                'end_line': 7,
            }
        ]
        self.assertRaises(IndexError, middleware._apply_dynamic_mappings)

    def test_invalid_dynamic_decorator_module(self):
        middleware = SilkyMiddleware(fake_get_response)
        SilkyConfig().SILKY_DYNAMIC_PROFILING = [
            {
                'module': 'tests.data.dfsdf',
                'function': 'foo'
            }
        ]
        self.assertRaises(AttributeError, middleware._apply_dynamic_mappings)

    def test_invalid_dynamic_decorator_function_name(self):
        middleware = SilkyMiddleware(fake_get_response)
        SilkyConfig().SILKY_DYNAMIC_PROFILING = [
            {
                'module': 'tests.data.dynamic',
                'function': 'bar'
            }
        ]
        self.assertRaises(AttributeError, middleware._apply_dynamic_mappings)

    def test_invalid_dynamic_mapping(self):
        middleware = SilkyMiddleware(fake_get_response)
        SilkyConfig().SILKY_DYNAMIC_PROFILING = [
            {
                'dfgdf': 'tests.data.dynamic',
                'funcgdfgtion': 'bar'
            }
        ]
        self.assertRaises(KeyError, middleware._apply_dynamic_mappings)

    def test_no_mappings(self):
        middleware = SilkyMiddleware(fake_get_response)
        SilkyConfig().SILKY_DYNAMIC_PROFILING = [

        ]
        middleware._apply_dynamic_mappings()  # Just checking no crash


class TestShouldIntercept(TestCase):
    def test_should_intercept_non_silk_request(self):
        request = Request()
        request.path = '/myapp/foo'
        should_intercept = _should_intercept(request)

        self.assertTrue(should_intercept)

    def test_should_intercept_silk_request(self):
        request = Request()
        request.path = reverse('silk:summary')
        should_intercept = _should_intercept(request)

        self.assertFalse(should_intercept)

    @override_settings(ROOT_URLCONF='tests.urlconf_without_silk')
    def test_should_intercept_without_silk_urls(self):
        request = Request()
        request.path = '/login'
        _should_intercept(request)  # Just checking no crash

    def test_should_intercept_ignore_paths(self):
        SilkyConfig().SILKY_IGNORE_PATHS = [
            '/ignorethis'
        ]
        request = Request()
        request.path = '/ignorethis'
        should_intercept = _should_intercept(request)

        self.assertFalse(should_intercept)



# Generated by Django 1.11.3 on 2017-07-31 23:40

from django.db import migrations, models

import silk.models


class Migration(migrations.Migration):

    dependencies = [
        ('silk', '0004_request_prof_file_storage'),
    ]

    operations = [
        migrations.AlterField(
            model_name='request',
            name='prof_file',
            field=models.FileField(max_length=300, null=True, storage=silk.models.silk_storage, upload_to=''),
        ),
    ]



from unittest.mock import patch

from django.test import TestCase

import silk
from silk.profiling.dynamic import (
    _get_module,
    _get_parent_module,
    profile_function_or_method,
)

from .test_lib.assertion import dict_contains
from .util import mock_data_collector


class TestGetModule(TestCase):
    """test for _get_module"""

    def test_singular(self):
        module = _get_module('silk')
        self.assertEqual(module.__class__.__name__, 'module')
        self.assertEqual('silk', module.__name__)
        self.assertTrue(hasattr(module, 'models'))

    def test_dot(self):
        module = _get_module('silk.models')
        self.assertEqual(module.__class__.__name__, 'module')
        self.assertEqual('silk.models', module.__name__)
        self.assertTrue(hasattr(module, 'SQLQuery'))


class TestGetParentModule(TestCase):
    """test for silk.tools._get_parent_module"""

    def test_singular(self):
        parent = _get_parent_module(silk)
        self.assertIsInstance(parent, dict)

    def test_dot(self):
        import silk.utils

        parent = _get_parent_module(silk.utils)
        self.assertEqual(parent, silk)


class MyClass:
    def foo(self):
        pass


def foo():
    pass


def source_file_name():
    file_name = __file__
    if file_name[-1] == 'c':
        file_name = file_name[:-1]
    return file_name


class TestProfileFunction(TestCase):
    def test_method_as_str(self):
        # noinspection PyShadowingNames
        def foo(_):
            pass

        # noinspection PyUnresolvedReferences
        with patch.object(MyClass, 'foo', foo):
            profile_function_or_method('tests.test_dynamic_profiling', 'MyClass.foo', 'test')
            dc = mock_data_collector()
            with patch('silk.profiling.profiler.DataCollector', return_value=dc) as mock_DataCollector:
                MyClass().foo()
                self.assertEqual(mock_DataCollector.return_value.register_profile.call_count, 1)
                call_args = mock_DataCollector.return_value.register_profile.call_args[0][0]
                self.assertTrue(dict_contains({
                    'func_name': foo.__name__,
                    'dynamic': True,
                    'file_path': source_file_name(),
                    'name': 'test',
                    'line_num': foo.__code__.co_firstlineno
                }, call_args))

    def test_func_as_str(self):
        name = foo.__name__
        line_num = foo.__code__.co_firstlineno
        profile_function_or_method('tests.test_dynamic_profiling', 'foo', 'test')
        dc = mock_data_collector()
        with patch('silk.profiling.profiler.DataCollector', return_value=dc) as mock_DataCollector:
            foo()
            self.assertEqual(mock_DataCollector.return_value.register_profile.call_count, 1)
            call_args = mock_DataCollector.return_value.register_profile.call_args[0][0]
            self.assertTrue(dict_contains({
                'func_name': name,
                'dynamic': True,
                'file_path': source_file_name(),
                'name': 'test',
                'line_num': line_num
            }, call_args))



from django.conf import settings
from django.db import connections


def delete_model(model):
    engine = settings.DATABASES[model.objects.db]['ENGINE']
    table = model._meta.db_table
    if 'mysql' in engine or 'postgresql' in engine:
        # Use "TRUNCATE" on the table
        with connections[model.objects.db].cursor() as cursor:
            if 'mysql' in engine:
                cursor.execute("SET FOREIGN_KEY_CHECKS=0;")
                cursor.execute(f"TRUNCATE TABLE {table}")
                cursor.execute("SET FOREIGN_KEY_CHECKS=1;")
            elif 'postgres' in engine:
                cursor.execute(f"ALTER TABLE {table} DISABLE TRIGGER USER;")
                cursor.execute(f"TRUNCATE TABLE {table} CASCADE")
                cursor.execute(f"ALTER TABLE {table} ENABLE TRIGGER USER;")
        return

    # Manually delete rows because sqlite does not support TRUNCATE and
    # oracle doesn't provide good support for disabling foreign key checks
    while True:
        items_to_delete = list(
            model.objects.values_list('pk', flat=True).all()[:800])
        if not items_to_delete:
            break
        model.objects.filter(pk__in=items_to_delete).delete()



from unittest.mock import Mock, NonCallableMagicMock, NonCallableMock, patch

from django.test import TestCase

from silk.collector import DataCollector
from silk.models import Request, SQLQuery
from silk.sql import execute_sql

from .util import delete_all_models


def mock_sql():
    mock_sql_query = Mock(spec_set=['_execute_sql', 'query', 'as_sql', 'connection'])
    mock_sql_query._execute_sql = Mock()
    mock_sql_query.query = NonCallableMock(spec_set=['model'])
    mock_sql_query.query.model = Mock()
    query_string = 'SELECT * from table_name'
    mock_sql_query.as_sql = Mock(return_value=(query_string, ()))

    mock_sql_query.connection = NonCallableMock(
        spec_set=['cursor', 'features', 'ops'],
        cursor=Mock(
            spec_set=['__call__'],
            return_value=NonCallableMagicMock(spec_set=['__enter__', '__exit__', 'execute'])
        ),
        features=NonCallableMock(
            spec_set=['supports_explaining_query_execution'],
            supports_explaining_query_execution=True
        ),
        ops=NonCallableMock(spec_set=['explain_query_prefix']),
    )

    return mock_sql_query, query_string


def call_execute_sql(cls, request):
    DataCollector().configure(request=request)
    delete_all_models(SQLQuery)
    cls.mock_sql, cls.query_string = mock_sql()
    kwargs = {
        'one': 1,
        'two': 2
    }
    cls.args = [1, 2]
    cls.kwargs = kwargs
    execute_sql(cls.mock_sql, *cls.args, **cls.kwargs)


class TestCallNoRequest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        call_execute_sql(cls, None)

    def test_called(self):
        self.mock_sql._execute_sql.assert_called_once_with(*self.args, **self.kwargs)

    def test_count(self):
        self.assertEqual(0, len(DataCollector().queries))


class TestCallRequest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        call_execute_sql(cls, Request())

    def test_called(self):
        self.mock_sql._execute_sql.assert_called_once_with(*self.args, **self.kwargs)

    def test_count(self):
        self.assertEqual(1, len(DataCollector().queries))

    def test_query(self):
        query = list(DataCollector().queries.values())[0]
        self.assertEqual(query['query'], self.query_string)


class TestCallSilky(TestCase):
    def test_no_effect(self):
        DataCollector().configure()
        sql, _ = mock_sql()
        sql.query.model = NonCallableMagicMock(spec_set=['__module__'])
        sql.query.model.__module__ = 'silk.models'
        # No SQLQuery models should be created for silk requests for obvious reasons
        with patch('silk.sql.DataCollector', return_value=Mock()) as mock_DataCollector:
            execute_sql(sql)
            self.assertFalse(mock_DataCollector().register_query.call_count)


class TestCollectorInteraction(TestCase):
    def _query(self):
        try:
            query = list(DataCollector().queries.values())[0]
        except IndexError:
            self.fail('No queries created')
        return query

    def test_request(self):
        DataCollector().configure(request=Request.objects.create(path='/path/to/somewhere'))
        sql, _ = mock_sql()
        execute_sql(sql)
        query = self._query()
        self.assertEqual(query['request'], DataCollector().request)

    def test_registration(self):
        DataCollector().configure(request=Request.objects.create(path='/path/to/somewhere'))
        sql, _ = mock_sql()
        execute_sql(sql)
        query = self._query()
        self.assertIn(query, DataCollector().queries.values())

    def test_explain(self):
        DataCollector().configure(request=Request.objects.create(path='/path/to/somewhere'))
        sql, qs = mock_sql()
        prefix = "EXPLAIN"
        mock_cursor = sql.connection.cursor.return_value.__enter__.return_value
        sql.connection.ops.explain_query_prefix.return_value = prefix
        execute_sql(sql)
        mock_cursor.execute.assert_called_once_with(f"{prefix} {qs}", ())



from django.contrib import admin
from django.urls import reverse

from .models import Blind


class BlindAdmin(admin.ModelAdmin):
    list_display = ('desc', 'thumbnail', 'name', 'child_safe')
    list_editable = ('name', 'child_safe')

    def thumbnail(self, obj):
        try:
            img_tag = '<img src="%s" width="200px"/>' % obj.photo.url
        except ValueError:
            return ''
        url = self._blind_url(obj)
        return f'<a href="{url}">{img_tag}</a>'

    thumbnail.short_description = 'Photo'
    thumbnail.allow_tags = True

    def _blind_url(self, obj):
        url = reverse('admin:example_app_blind_change', args=(obj.id, ))
        return url

    def desc(self, obj):
        desc = str(obj)
        url = self._blind_url(obj)
        return f'<a href="{url}">{desc}</a>'

    desc.short_description = 'Blind'
    desc.allow_tags = True


admin.site.register(Blind, BlindAdmin)



from django.db.models import Sum
from django.shortcuts import render
from django.template.context_processors import csrf
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.models import Request, Response
from silk.request_filters import BaseFilter, filters_from_request

__author__ = 'mtford'


class RequestsView(View):
    show = [5, 10, 25, 100, 250]
    default_show = 25

    order_by = {
        'start_time': {
            'label': 'Recent',
            'additional_query_filter': None
        },
        'path': {
            'label': 'Path',
            'additional_query_filter': None
        },
        'num_sql_queries': {
            'label': 'Num. Queries',
            'additional_query_filter': None
        },
        'time_taken': {
            'label': 'Time',
            'additional_query_filter': lambda x: x.filter(time_taken__gte=0)
        },
        'db_time': {
            'label': 'Time on queries',
            'additional_query_filter': lambda x: x.annotate(db_time=Sum('queries__time_taken'))
            .filter(db_time__gte=0)
        },
    }
    order_dir = {
        'ASC': {
            'label': 'Ascending'
        },
        'DESC': {
            'label': 'Descending'
        }
    }
    view_style = {
        'card': {
            'label': 'Cards'
        },
        'row': {
            'label': 'Rows'
        }
    }
    default_order_by = 'start_time'
    default_order_dir = 'DESC'
    default_view_style = 'card'

    session_key_request_filters = 'request_filters'

    @property
    def options_order_by(self):
        return [{'value': x, 'label': self.order_by[x]['label']} for x in self.order_by.keys()]

    @property
    def options_order_dir(self):
        return [{'value': x, 'label': self.order_dir[x]['label']} for x in self.order_dir.keys()]

    @property
    def options_view_style(self):
        return [{'value': x, 'label': self.view_style[x]['label']} for x in self.view_style.keys()]

    def _get_paths(self):
        return Request.objects.values_list(
            'path',
            flat=True
        ).order_by(
            'path'
        ).distinct()

    def _get_views(self):
        return Request.objects.values_list(
            'view_name',
            flat=True
        ).exclude(
            view_name=''
        ).order_by(
            'view_name'
        ).distinct()

    def _get_status_codes(self):
        return Response.objects.values_list(
            'status_code',
            flat=True
        ).order_by(
            'status_code'
        ).distinct()

    def _get_methods(self):
        return Request.objects.values_list(
            'method',
            flat=True
        ).order_by(
            'method'
        ).distinct()

    def _get_objects(self, show=None, order_by=None, order_dir=None, path=None, filters=None):
        if not filters:
            filters = []
        if not show:
            show = self.default_show
        query_set = Request.objects.all()
        if not order_by:
            order_by = self.default_order_by
        if not order_dir:
            order_dir = self.default_order_dir
        if order_by not in self.order_by.keys():
            raise RuntimeError('Unknown order_by: "%s"' % order_by)
        ob = self.order_by[order_by]
        if ob['additional_query_filter'] is not None:
            query_set = ob['additional_query_filter'](query_set)
        query_set = query_set.order_by('{}{}'.format('-' if order_dir == 'DESC' else '', order_by))
        if path:
            query_set = query_set.filter(path=path)
        for f in filters:
            query_set = f.contribute_to_query_set(query_set)
            query_set = query_set.filter(f)
        return query_set[:show]

    def _create_context(self, request):
        show = request.GET.get('show', self.default_show)
        order_by = request.GET.get('order_by', self.default_order_by)
        order_dir = request.GET.get('order_dir', self.default_order_dir)
        view_style = request.GET.get('view_style', self.default_view_style)

        if show:
            show = int(show)
        path = request.GET.get('path', None)
        raw_filters = request.session.get(self.session_key_request_filters, {})
        context = {
            'show': show,
            'order_by': order_by,
            'order_dir': order_dir,
            'view_style': view_style,
            'request': request,
            'options_show': self.show,
            'options_order_by': self.options_order_by,
            'options_order_dir': self.options_order_dir,
            'options_view_style': self.options_view_style,
            'options_paths': self._get_paths(),
            'options_status_codes': self._get_status_codes(),
            'options_methods': self._get_methods(),
            'view_names': self._get_views(),
            'filters': raw_filters
        }
        context.update(csrf(request))
        if path:
            context['path'] = path
        context['results'] = self._get_objects(show, order_by, order_dir, path,
                                               filters=[BaseFilter.from_dict(x) for _, x in raw_filters.items()])
        return context

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request):
        return render(request, 'silk/requests.html', self._create_context(request))

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def post(self, request):
        filters = filters_from_request(request)
        request.session[self.session_key_request_filters] = {ident: f.as_dict() for ident, f in filters.items()}
        return render(request, 'silk/requests.html', self._create_context(request))



def dict_contains(child_dict, parent_dict):
    for key, value in child_dict.items():
        if key not in parent_dict:
            return False
        if parent_dict[key] != value:
            return False
    return True



from time import sleep

from django.test import TestCase

from silk.collector import DataCollector
from silk.models import Request, _time_taken
from silk.profiling.profiler import silk_profile

from .test_lib.mock_suite import MockSuite


class TestProfilerRequests(TestCase):
    def test_context_manager_no_request(self):
        DataCollector().configure()
        with silk_profile(name='test_profile'):
            sleep(0.1)
        self.assertFalse(DataCollector().profiles)

    def test_decorator_no_request(self):
        DataCollector().configure()

        @silk_profile()
        def func():
            sleep(0.1)

        func()
        profile = list(DataCollector().profiles.values())[0]
        self.assertFalse(profile['request'])

    def test_context_manager_request(self):
        DataCollector().configure(Request.objects.create(path='/to/somewhere'))
        with silk_profile(name='test_profile'):
            sleep(0.1)
        profile = list(DataCollector().profiles.values())[0]
        self.assertEqual(DataCollector().request, profile['request'])

    def test_decorator_request(self):
        DataCollector().configure(Request.objects.create(path='/to/somewhere'))

        @silk_profile()
        def func():
            sleep(0.1)

        func()
        profile = list(DataCollector().profiles.values())[0]
        self.assertEqual(DataCollector().request, profile['request'])


class TestProfilertContextManager(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        r = Request.objects.create()
        DataCollector().configure(r)
        with silk_profile(name='test_profile'):
            sleep(0.1)

    def test_one_object(self):
        self.assertEqual(len(DataCollector().profiles), 1)

    def test_name(self):
        profile = list(DataCollector().profiles.values())[0]
        self.assertEqual(profile['name'], 'test_profile')

    def test_time_taken(self):
        profile = list(DataCollector().profiles.values())[0]
        time_taken = _time_taken(start_time=profile['start_time'], end_time=profile['end_time'])
        self.assertGreaterEqual(time_taken, 100)
        self.assertLess(time_taken, 110)


class TestProfilerDecorator(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        DataCollector().configure(Request.objects.create())

        @silk_profile()
        def func():
            sleep(0.1)

        func()

    def test_one_object(self):
        self.assertEqual(len(DataCollector().profiles), 1)

    def test_name(self):
        profile = list(DataCollector().profiles.values())[0]
        self.assertEqual(profile['name'], 'func')

    def test_time_taken(self):
        profile = list(DataCollector().profiles.values())[0]
        time_taken = _time_taken(start_time=profile['start_time'], end_time=profile['end_time'])
        self.assertGreaterEqual(time_taken, 100)
        self.assertLess(time_taken, 115)


class TestQueries(TestCase):
    def test_no_queries_before(self):
        DataCollector().configure(Request.objects.create())
        with silk_profile(name='test_no_queries_before_profile'):
            mock_queries = MockSuite().mock_sql_queries(n=5, as_dict=True)
            DataCollector().register_query(*mock_queries)
        profile = list(DataCollector().profiles.values())[0]
        self.assertEqual(profile['name'], 'test_no_queries_before_profile')
        queries = profile['queries']
        self.assertEqual(len(queries), 5)
        for query in DataCollector().queries:
            self.assertIn(query, queries)

    def test_queries_before(self):
        """test that any queries registered before profiling begins are ignored"""
        DataCollector().configure(Request.objects.create())
        DataCollector().register_query(*MockSuite().mock_sql_queries(n=2, as_dict=True))
        before = [x for x in DataCollector().queries]
        with silk_profile(name='test_no_queries_before_profile'):
            mock_queries = MockSuite().mock_sql_queries(n=5, as_dict=True)
            DataCollector().register_query(*mock_queries)
        profile = list(DataCollector().profiles.values())[0]
        self.assertEqual(profile['name'], 'test_no_queries_before_profile')
        queries = profile['queries']
        self.assertEqual(len(queries), 5)
        for query in set(DataCollector().queries).difference(before):
            self.assertIn(query, queries)



from django.http import FileResponse
from django.shortcuts import get_object_or_404
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.models import Request


class ProfileDownloadView(View):

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, request_id):
        silk_request = get_object_or_404(Request, pk=request_id, prof_file__isnull=False)
        response = FileResponse(silk_request.prof_file)
        response['Content-Disposition'] = f'attachment; filename="{silk_request.prof_file.name}"'
        return response



from django.db.models import Avg, Count, Max, Sum
from django.shortcuts import render
from django.template.context_processors import csrf
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk import models
from silk.auth import login_possibly_required, permissions_possibly_required
from silk.request_filters import BaseFilter, filters_from_request


class SummaryView(View):
    filters_key = 'summary_filters'

    def _avg_num_queries(self, filters):
        queries__aggregate = models.Request.objects.filter(*filters).annotate(num_queries=Count('queries')).aggregate(num=Avg('num_queries'))
        return queries__aggregate['num']

    def _avg_time_spent_on_queries(self, filters):
        taken__aggregate = models.Request.objects.filter(*filters).annotate(time_spent=Sum('queries__time_taken')).aggregate(num=Avg('time_spent'))
        return taken__aggregate['num']

    def _avg_overall_time(self, filters):
        taken__aggregate = models.Request.objects.filter(*filters).annotate(time_spent=Sum('time_taken')).aggregate(num=Avg('time_spent'))
        return taken__aggregate['num']

    # TODO: Find a more efficient way to do this. Currently has to go to DB num. views + 1 times and is prob quite expensive
    def _longest_query_by_view(self, filters):
        values_list = models.Request.objects.filter(*filters).values_list("view_name").annotate(max=Max('time_taken')).filter(max__isnull=False).order_by('-max')[:5]
        requests = []
        for view_name, _ in values_list:
            request = models.Request.objects.filter(view_name=view_name, *filters).filter(time_taken__isnull=False).order_by('-time_taken')[0]
            requests.append(request)
        return sorted(requests, key=lambda item: item.time_taken, reverse=True)

    def _time_spent_in_db_by_view(self, filters):
        values_list = models.Request.objects.filter(*filters).values_list('view_name').annotate(t=Sum('queries__time_taken')).filter(t__gte=0).order_by('-t')[:5]
        requests = []
        for view, _ in values_list:
            r = models.Request.objects.filter(view_name=view, *filters).annotate(t=Sum('queries__time_taken')).filter(t__isnull=False).order_by('-t')[0]
            requests.append(r)
        return sorted(requests, key=lambda item: item.t, reverse=True)

    def _num_queries_by_view(self, filters):
        queryset = models.Request.objects.filter(*filters).values_list('view_name').annotate(t=Count('queries')).order_by('-t')[:5]
        views = [r[0] for r in queryset[:6]]
        requests = []
        for view in views:
            try:
                r = models.Request.objects.filter(view_name=view, *filters).annotate(t=Count('queries')).order_by('-t')[0]
                requests.append(r)
            except IndexError:
                pass
        return sorted(requests, key=lambda item: item.t, reverse=True)

    def _create_context(self, request):
        raw_filters = request.session.get(self.filters_key, {})
        filters = [BaseFilter.from_dict(filter_d) for _, filter_d in raw_filters.items()]
        avg_overall_time = self._avg_num_queries(filters)
        c = {
            'request': request,
            'num_requests': models.Request.objects.filter(*filters).count(),
            'num_profiles': models.Profile.objects.filter(*filters).count(),
            'avg_num_queries': avg_overall_time,
            'avg_time_spent_on_queries': self._avg_time_spent_on_queries(filters),
            'avg_overall_time': self._avg_overall_time(filters),
            'longest_queries_by_view': self._longest_query_by_view(filters),
            'most_time_spent_in_db': self._time_spent_in_db_by_view(filters),
            'most_queries': self._num_queries_by_view(filters),
            'filters': raw_filters
        }
        c.update(csrf(request))
        return c

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request):
        c = self._create_context(request)
        return render(request, 'silk/summary.html', c)

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def post(self, request):
        filters = filters_from_request(request)
        request.session[self.filters_key] = {ident: f.as_dict() for ident, f in filters.items()}
        return render(request, 'silk/summary.html', self._create_context(request))



# Generated by Django 1.9.7 on 2016-07-08 18:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('silk', '0002_auto_update_uuid4_id_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='request',
            name='prof_file',
            field=models.FileField(null=True, upload_to=''),
        ),
    ]



# Generated by Django 1.10.4 on 2016-12-06 00:23

from django.db import migrations, models

import silk.models


class Migration(migrations.Migration):

    dependencies = [
        ('silk', '0003_request_prof_file'),
    ]

    operations = [
        migrations.AlterField(
            model_name='request',
            name='prof_file',
            field=models.FileField(null=True, storage=silk.models.silk_storage, upload_to=''),
        ),
    ]



import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('silk', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='request',
            name='id',
            field=models.CharField(default=uuid.uuid4, max_length=36, serialize=False, primary_key=True),
        ),
        migrations.AlterField(
            model_name='response',
            name='id',
            field=models.CharField(default=uuid.uuid4, max_length=36, serialize=False, primary_key=True),
        ),
    ]



"""
Django queryset filters used by the requests view
"""
import logging
from datetime import datetime, timedelta

from django.db.models import Count, Q, Sum
from django.utils import timezone

from silk.profiling.dynamic import _get_module
from silk.templatetags.silk_filters import _silk_date_time

logger = logging.getLogger('silk.request_filters')


class FilterValidationError(Exception):
    pass


class BaseFilter(Q):
    def __init__(self, value=None, *args, **kwargs):
        self.value = value
        super().__init__(*args, **kwargs)

    @property
    def typ(self):
        return self.__class__.__name__

    @property
    def serialisable_value(self):
        return self.value

    def as_dict(self):
        return {'typ': self.typ, 'value': self.serialisable_value, 'str': str(self)}

    @staticmethod
    def from_dict(d):
        typ = d['typ']
        filter_class = globals()[typ]
        val = d.get('value', None)
        return filter_class(val)

    def contribute_to_query_set(self, query_set):
        """
        make any changes to the query-set before the query is applied,
        e.g. annotate with extra fields
        :param query_set: a django queryset
        :return: a new query set that this filter can then be used with
        """
        return query_set


class SecondsFilter(BaseFilter):
    def __init__(self, n):
        if n:
            try:
                value = int(n)
            except ValueError as e:
                raise FilterValidationError(e)
            now = timezone.now()
            frm_dt = now - timedelta(seconds=value)
            super().__init__(value, start_time__gt=frm_dt)
        else:
            # Empty query
            super().__init__()

    def __str__(self):
        return '>%d seconds ago' % self.value


def _parse(dt, fmt):
    """attempt to coerce dt into a datetime given fmt, otherwise raise
    a FilterValidationError"""
    try:
        dt = datetime.strptime(dt, fmt)
    except TypeError:
        if not isinstance(dt, datetime):
            raise FilterValidationError('Must be a datetime object')
    except ValueError as e:
        raise FilterValidationError(e)
    return dt


class BeforeDateFilter(BaseFilter):
    fmt = '%Y/%m/%d %H:%M'

    def __init__(self, dt):
        value = _parse(dt, self.fmt)
        super().__init__(value, start_time__lt=value)

    @property
    def serialisable_value(self):
        return self.value.strftime(self.fmt)

    def __str__(self):
        return '<%s' % _silk_date_time(self.value)


class AfterDateFilter(BaseFilter):
    fmt = '%Y/%m/%d %H:%M'

    def __init__(self, dt):
        value = _parse(dt, self.fmt)
        super().__init__(value, start_time__gt=value)

    @property
    def serialisable_value(self):
        return self.value.strftime(self.fmt)

    def __str__(self):
        return '>%s' % _silk_date_time(self.value)


class ViewNameFilter(BaseFilter):
    """filter on the name of the view, e.g. the name=xyz component of include in urls.py"""

    def __init__(self, view_name):
        value = view_name
        super().__init__(value, view_name=view_name)

    def __str__(self):
        return 'View == %s' % self.value


class PathFilter(BaseFilter):
    """filter on path e.g. /path/to/something"""

    def __init__(self, path):
        value = path
        super().__init__(value, path=path)

    def __str__(self):
        return 'Path == %s' % self.value


class NameFilter(BaseFilter):
    def __init__(self, name):
        value = name
        super().__init__(value, name=name)

    def __str__(self):
        return 'name == %s' % self.value


class FunctionNameFilter(BaseFilter):
    def __init__(self, func_name):
        value = func_name
        super().__init__(value, func_name=func_name)

    def __str__(self):
        return 'func_name == %s' % self.value


class NumQueriesFilter(BaseFilter):
    def __init__(self, n):
        try:
            value = int(n)
        except ValueError as e:
            raise FilterValidationError(e)
        super().__init__(value, num_queries__gte=n)

    def __str__(self):
        return '#queries >= %s' % self.value

    def contribute_to_query_set(self, query_set):
        return query_set.annotate(num_queries=Count('queries'))


class TimeSpentOnQueriesFilter(BaseFilter):
    def __init__(self, n):
        try:
            value = int(n)
        except ValueError as e:
            raise FilterValidationError(e)
        super().__init__(value, db_time__gte=n)

    def __str__(self):
        return 'DB Time >= %s' % self.value

    def contribute_to_query_set(self, query_set):
        return query_set.annotate(db_time=Sum('queries__time_taken'))


class OverallTimeFilter(BaseFilter):
    def __init__(self, n):
        try:
            value = int(n)
        except ValueError as e:
            raise FilterValidationError(e)
        super().__init__(value, time_taken__gte=n)

    def __str__(self):
        return 'Time >= %s' % self.value


class StatusCodeFilter(BaseFilter):
    def __init__(self, n):
        try:
            value = int(n)
        except ValueError as e:
            raise FilterValidationError(e)
        super().__init__(value, response__status_code=n)


class MethodFilter(BaseFilter):
    def __init__(self, value):
        super().__init__(value, method=value)


def filters_from_request(request):
    raw_filters = {}
    for key in request.POST:
        splt = key.split('-')
        if splt[0].startswith('filter'):
            ident = splt[1]
            typ = splt[2]
            if ident not in raw_filters:
                raw_filters[ident] = {}
            raw_filters[ident][typ] = request.POST[key]
    filters = {}
    for ident, raw_filter in raw_filters.items():
        value = raw_filter.get('value', '')
        if value.strip():
            typ = raw_filter['typ']
            module = _get_module('silk.request_filters')
            filter_class = getattr(module, typ)
            try:
                f = filter_class(value)
                filters[ident] = f
            except FilterValidationError:
                logger.warning(f'Validation error when processing filter {typ}({value})')
    return filters



import json
from unittest.mock import Mock

from django.test import TestCase

from silk.config import SilkyConfig
from silk.model_factory import RequestModelFactory

DJANGO_META_CONTENT_TYPE = 'CONTENT_TYPE'
HTTP_CONTENT_TYPE = 'Content-Type'
CLEANSED = RequestModelFactory.CLEANSED_SUBSTITUTE
DEFAULT_SENSITIVE_KEYS = {'username', 'api', 'token', 'key', 'secret', 'password', 'signature'}


class MaskCredentialsInFormsTest(TestCase):
    def tearDown(self):
        SilkyConfig().SILKY_SENSITIVE_KEYS = DEFAULT_SENSITIVE_KEYS

    def _mask(self, value):
        return RequestModelFactory(None)._mask_credentials(value)

    def test_mask_credentials_preserves_single_insensitive_values(self):
        body = "foo=public"
        expected = "foo=public"
        self.assertEqual(expected, self._mask(body))

    def test_mask_credentials_masks_sensitive_values(self):
        body = "password=secret"
        expected = f"password={CLEANSED}"
        self.assertEqual(expected, self._mask(body))

    def test_mask_credentials_masks_multiple_sensitive_values(self):
        body = "password=mypassword&secret=mysecret"
        expected = f"password={CLEANSED}&secret={CLEANSED}"
        self.assertEqual(expected, self._mask(body))

    def test_mask_credentials_masks_sensitive_values_between_insensitive_values(self):
        body = "public1=foo&password=secret&public2=bar"
        expected = f"public1=foo&password={CLEANSED}&public2=bar"
        self.assertEqual(expected, self._mask(body))

    def test_mask_credentials_preserves_insensitive_values_between_sensitive_values(self):
        body = "password=1&foo=public&secret=2"
        expected = f"password={CLEANSED}&foo=public&secret={CLEANSED}"
        self.assertEqual(expected, self._mask(body))

    def test_mask_credentials_is_case_insensitive(self):
        body = "UsErNaMe=secret"
        expected = f"UsErNaMe={CLEANSED}"
        self.assertEqual(expected, self._mask(body))

    def test_mask_credentials_handles_prefixes(self):
        body = "prefixed-username=secret"
        expected = f"prefixed-username={CLEANSED}"
        self.assertEqual(expected, self._mask(body))

    def test_mask_credentials_handles_suffixes(self):
        body = "username-with-suffix=secret"
        expected = f"username-with-suffix={CLEANSED}"
        self.assertEqual(expected, self._mask(body))

    def test_mask_credentials_handles_regex_characters(self):
        body = "password=secret++"
        expected = f"password={CLEANSED}"
        self.assertEqual(expected, self._mask(body))

    def test_mask_credentials_handles_complex_cases(self):
        body = "foo=public&prefixed-uSeRname-with-suffix=secret&bar=public"
        expected = f"foo=public&prefixed-uSeRname-with-suffix={CLEANSED}&bar=public"
        self.assertEqual(expected, self._mask(body))

    def test_mask_credentials_masks_sensitive_values_listed_in_settings(self):
        SilkyConfig().SILKY_SENSITIVE_KEYS = {"foo"}
        body = "foo=hidethis"
        expected = f"foo={CLEANSED}"
        self.assertEqual(expected, self._mask(body))


class MaskCredentialsInJsonTest(TestCase):
    def tearDown(self):
        SilkyConfig().SILKY_SENSITIVE_KEYS = DEFAULT_SENSITIVE_KEYS

    def _mask(self, value):
        return RequestModelFactory(None)._mask_credentials(json.dumps(value))

    def test_mask_credentials_preserves_single_insensitive_values(self):
        self.assertIn("public", self._mask({"foo": "public"}))

    def test_mask_credentials_preserves_insensitive_values_in_presence_of_sensitive(self):
        self.assertIn("public", self._mask({"password": "secret", "foo": "public"}))

    def test_mask_credentials_masks_sensitive_values(self):
        self.assertNotIn("secret", self._mask({"password": "secret"}))

    def test_mask_credentials_masks_sensitive_values_in_presence_of_regular(self):
        self.assertNotIn("secret", self._mask({"foo": "public", "password": "secret"}))

    def test_mask_credentials_is_case_insensitive(self):
        self.assertNotIn("secret", self._mask({"UsErNaMe": "secret"}))

    def test_mask_credentials_handles_prefixes(self):
        self.assertNotIn("secret", self._mask({"prefixed-username": "secret"}))

    def test_mask_credentials_handles_suffixes(self):
        self.assertNotIn("secret", self._mask({"username-with-suffix": "secret"}))

    def test_mask_credentials_handles_complex_cases(self):
        self.assertNotIn("secret", self._mask({
            "foo": "public",
            "prefixed-uSeRname-with-suffix": "secret"
        }))

    def test_mask_credentials_in_nested_data_structures(self):
        self.assertNotIn("secret", self._mask({
            "foo": "public",
            "nested": {
                "prefixed-uSeRname-with-suffix": "secret",
            },
        }))

    def test_mask_credentials_masks_sensitive_values_listed_in_settings(self):
        SilkyConfig().SILKY_SENSITIVE_KEYS = {"foo"}
        self.assertNotIn("hidethis", self._mask({"foo": "hidethis"}))


class TestEncodingForRequests(TestCase):
    """
    Check that the RequestModelFactory masks sensitive data
    """

    def test_password_in_body(self):
        mock_request = Mock()
        mock_request.META = {DJANGO_META_CONTENT_TYPE: 'text/plain'}
        mock_request.body = 'username=test_username&unmasked=testunmasked&password=testpassword'
        mock_request.get = mock_request.META.get
        factory = RequestModelFactory(mock_request)
        body, raw_body = factory.body()

        self.assertIn('testunmasked', raw_body)
        self.assertNotIn('test_username', raw_body)
        self.assertNotIn('testpassword', raw_body)
        self.assertNotIn('test_username', body)
        self.assertNotIn('testpassword', body)

    def test_password_in_json(self):
        mock_request = Mock()
        mock_request.META = {DJANGO_META_CONTENT_TYPE: 'application/json; charset=UTF-8'}
        d = {'x': 'testunmasked', 'username': 'test_username', 'password': 'testpassword',
             'prefixed-secret': 'testsecret'}
        mock_request.body = json.dumps(d)
        mock_request.get = mock_request.META.get
        factory = RequestModelFactory(mock_request)
        body, raw_body = factory.body()
        self.assertIn('testunmasked', raw_body)
        self.assertNotIn('test_username', raw_body)
        self.assertNotIn('testpassword', raw_body)
        self.assertNotIn('testsecret', raw_body)
        self.assertNotIn('test_username', body)
        self.assertNotIn('testpassword', body)
        self.assertNotIn('testsecret', body)

        for datum in [json.loads(body), json.loads(raw_body)]:
            self.assertEqual(datum['username'], RequestModelFactory.CLEANSED_SUBSTITUTE)
            self.assertEqual(datum['password'], RequestModelFactory.CLEANSED_SUBSTITUTE)
            self.assertEqual(datum['prefixed-secret'], RequestModelFactory.CLEANSED_SUBSTITUTE)
            self.assertEqual(datum['x'], 'testunmasked')

    def test_password_in_batched_json(self):
        mock_request = Mock()
        mock_request.META = {DJANGO_META_CONTENT_TYPE: 'application/json; charset=UTF-8'}
        d = [
            {'x': 'testunmasked', 'username': 'test_username', 'password': 'testpassword'},
            {'x': 'testunmasked', 'username': 'test_username', 'password': 'testpassword'}
        ]
        mock_request.body = json.dumps(d)
        mock_request.get = mock_request.META.get
        factory = RequestModelFactory(mock_request)
        body, raw_body = factory.body()
        self.assertIn('testunmasked', raw_body)
        self.assertNotIn('test_username', raw_body)
        self.assertNotIn('testpassword', raw_body)
        self.assertNotIn('test_username', body[0])
        self.assertNotIn('testpassword', body[0])
        self.assertNotIn('test_username', body[1])
        self.assertNotIn('testpassword', body[1])

        for data in [json.loads(body), json.loads(raw_body)]:
            for datum in data:
                self.assertEqual(datum['username'], RequestModelFactory.CLEANSED_SUBSTITUTE)
                self.assertEqual(datum['password'], RequestModelFactory.CLEANSED_SUBSTITUTE)
                self.assertEqual(datum['x'], 'testunmasked')

    def test_authorization_header(self):
        mock_request = Mock()
        mock_request.META = {'HTTP_AUTHORIZATION': 'secret'}
        mock_request.body = ''
        mock_request.get = mock_request.META.get
        factory = RequestModelFactory(mock_request)
        headers = factory.encoded_headers()
        json_headers = json.loads(headers)

        self.assertIn('AUTHORIZATION', json_headers)
        self.assertEqual(json_headers['AUTHORIZATION'], RequestModelFactory.CLEANSED_SUBSTITUTE)



import json

# noinspection PyUnresolvedReferences
from urllib.parse import urlencode

import jinja2

curl_template = """
curl {% if method %}-X {{ method }}{% endif %}
{% if content_type %}-H 'Content-Type: {{ content_type }}'{% endif %}
{% if modifier %}{{ modifier }} {% endif %}{% if body %}'{{ body }}'{% endif %}
{{ url }}{% if query_params %}{{ query_params }}{% endif %}
{% if extra %}{{ extra }}{% endif %}
"""


def _curl_process_params(body, content_type, query_params):
    extra = None
    modifier = None
    if query_params:
        try:
            query_params = urlencode(
                [(k, v.encode('utf8')) for k, v in query_params.items()]
            )
        except TypeError:
            pass
        query_params = '?' + str(query_params)
    if 'json' in content_type or 'javascript' in content_type:
        if isinstance(body, dict):
            body = json.dumps(body)
        modifier = '-d'
    # See http://curl.haxx.se/docs/manpage.html#-F
    # for multipart vs x-www-form-urlencoded
    # x-www-form-urlencoded is same way as browser,
    # multipart is RFC 2388 which allows file uploads.
    elif 'multipart' in content_type or 'x-www-form-urlencoded' in content_type:
        try:
            body = ' '.join([f'{k}={v}' for k, v in body.items()])
        except AttributeError:
            modifier = '-d'
        else:
            content_type = None
            modifier = '-F'
    elif body:
        body = str(body)
        modifier = '-d'
    else:
        modifier = None
        content_type = None
    # TODO: Clean up.
    return modifier, body, query_params, content_type, extra


def curl_cmd(url, method=None, query_params=None, body=None, content_type=None):
    if not content_type:
        content_type = 'text/plain'
    modifier, body, query_params, content_type, extra = _curl_process_params(
        body, content_type, query_params
    )
    t = jinja2.Template(curl_template)
    return ' '.join(t.render(url=url,
                             method=method,
                             query_params=query_params,
                             body=body,
                             modifier=modifier,
                             content_type=content_type,
                             extra=extra).split('\n'))



from django.core.files.storage import FileSystemStorage

from silk.config import SilkyConfig


class ProfilerResultStorage(FileSystemStorage):
    # the default storage will only store under MEDIA_ROOT, so we must define our own.
    def __init__(self):
        super().__init__(
            location=SilkyConfig().SILKY_PYTHON_PROFILER_RESULT_PATH,
            base_url=''
        )
        self.base_url = None



import json
from unittest.mock import Mock

from django.test import TestCase

from silk.model_factory import ResponseModelFactory

DJANGO_META_CONTENT_TYPE = 'CONTENT_TYPE'
HTTP_CONTENT_TYPE = 'Content-Type'


class TestByteStringCompatForResponse(TestCase):

    def test_bytes_compat(self):
        """
        Test ResponseModelFactory formats json with bytes content
        """
        mock = Mock()
        mock.headers = {HTTP_CONTENT_TYPE: 'application/json;'}
        d = {'k': 'v'}
        mock.content = bytes(json.dumps(d), 'utf-8')
        mock.get = mock.headers.get
        factory = ResponseModelFactory(mock)
        body, content = factory.body()
        self.assertDictEqual(json.loads(body), d)

    # Testing invalid json throws an exception in the tests
    # def test_python3_invalid_content_compat(self):
    #     """
    #     Test ResponseModelFactory returns empty string for invalid json
    #     """
    #     if sys.version_info >= (3, 0, 0):
    #         mock = Mock()
    #         mock.pk = 'test'
    #         mock._headers = {HTTP_CONTENT_TYPE: 'application/json;'}
    #         mock.content = b'invalid json'
    #         mock.get = mock._headers.get
    #         factory = ResponseModelFactory(mock)
    #         body, content = factory.body()
    #         self.assertEqual(body, '')



from functools import WRAPPER_ASSIGNMENTS, wraps

from django.contrib.auth.decorators import login_required
from django.core.exceptions import PermissionDenied

from silk.config import SilkyConfig


def login_possibly_required(function=None, **kwargs):
    if SilkyConfig().SILKY_AUTHENTICATION:
        return login_required(function, **kwargs)
    return function


def permissions_possibly_required(function=None):
    if SilkyConfig().SILKY_AUTHORISATION:
        actual_decorator = user_passes_test(
            SilkyConfig().SILKY_PERMISSIONS
        )
        if function:
            return actual_decorator(function)
        return actual_decorator
    return function


def user_passes_test(test_func):
    def decorator(view_func):
        @wraps(view_func, assigned=WRAPPER_ASSIGNMENTS)
        def _wrapped_view(request, *args, **kwargs):
            if test_func(request.user):
                return view_func(request, *args, **kwargs)
            else:
                raise PermissionDenied

        return _wrapped_view

    return decorator



from unittest.mock import Mock

from django.test import TestCase

from silk.views.profiling import ProfilingView

from .test_lib.assertion import dict_contains
from .test_lib.mock_suite import MockSuite


class TestProfilingViewDefaults(TestCase):
    def test_func_names(self):
        profiles = [MockSuite().mock_profile() for _ in range(0, 3)]
        func_names = ProfilingView()._get_function_names()
        for p in profiles:
            self.assertIn(p.func_name, func_names)
        self.assertIn('', func_names)

    def test_show(self):
        self.assertIn(ProfilingView.default_show, ProfilingView.show)

    def test_order_by(self):
        self.assertIn(ProfilingView.defualt_order_by, ProfilingView.order_by)


class TestProfilingViewGetObjects(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.profiles = [MockSuite().mock_profile() for _ in range(0, 10)]

    def test_ordering(self):
        results = ProfilingView()._get_objects(order_by='Recent')
        self.assertSorted(results, 'start_time')

    def test_show(self):
        results = ProfilingView()._get_objects(show=5)
        self.assertEqual(5, len(results))

    def test_func_name(self):
        func_name = 'a_func_name'
        self.profiles[1].func_name = func_name
        self.profiles[1].save()
        results = ProfilingView()._get_objects(func_name=func_name)
        for r in results:
            self.assertEqual(r.func_name, func_name)

    def assertSorted(self, objects, sort_field):
        for idx, r in enumerate(objects):
            try:
                nxt = objects[idx + 1]
                self.assertGreaterEqual(getattr(r, sort_field), getattr(nxt, sort_field))
            except IndexError:
                pass


class TestProfilingContext(TestCase):
    def test_default(self):
        request = Mock(spec_set=['GET', 'session'])
        request.GET = {}
        request.session = {}
        context = ProfilingView()._create_context(request)
        self.assertTrue(dict_contains({
            'show': ProfilingView.default_show,
            'order_by': ProfilingView.defualt_order_by,
            'options_show': ProfilingView.show,
            'options_order_by': ProfilingView.order_by,
            'options_func_names': ProfilingView()._get_function_names()
        }, context))
        self.assertNotIn('path', context)
        self.assertIn('results', context)

    def test_get(self):
        request = Mock(spec_set=['GET', 'session'])
        request.session = {}
        show = 10
        func_name = 'func_name'
        name = 'name'
        order_by = 'Time'
        request.GET = {'show': show,
                       'func_name': func_name,
                       'name': name,
                       'order_by': order_by}
        context = ProfilingView()._create_context(request)
        self.assertTrue(dict_contains({
            'show': show,
            'order_by': order_by,
            'func_name': func_name,
            'name': name,
            'options_show': ProfilingView.show,
            'options_order_by': ProfilingView.order_by,
            'options_func_names': ProfilingView()._get_function_names()
        }, context))
        self.assertIn('results', context)



from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.models import Profile
from silk.views.code import _code_context, _code_context_from_request


class ProfilingDetailView(View):

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, *_, **kwargs):
        profile_id = kwargs['profile_id']
        context = {
            'request': request
        }
        profile = Profile.objects.get(pk=profile_id)
        file_path = profile.file_path
        line_num = profile.line_num

        context['pos'] = pos = int(request.GET.get('pos', 0))
        if pos:
            context.update(_code_context_from_request(request, prefix='pyprofile_'))

        context['profile'] = profile
        context['line_num'] = file_path
        context['file_path'] = line_num
        context['file_column'] = 5

        if profile.request:
            context['silk_request'] = profile.request
        if file_path and line_num:
            try:
                context.update(_code_context(file_path, line_num, profile.end_line_num))
            except OSError as e:
                if e.errno == 2:
                    context['code_error'] = e.filename + ' does not exist.'
                else:
                    raise e

        return render(request, 'silk/profile_detail.html', context)



# std
import json
import os
import shutil
import tempfile
from contextlib import closing, contextmanager

# 3rd party
from io import StringIO

from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.utils.decorators import method_decorator
from django.views.generic import View
from gprof2dot import DotWriter, PstatsParser, Theme

# silk
from silk.auth import login_possibly_required, permissions_possibly_required
from silk.models import Request

COLOR_MAP = Theme(
    mincolor=(0.18, 0.51, 0.53),
    maxcolor=(0.03, 0.49, 0.50),
    gamma=1.5,
    fontname='FiraSans',
    minfontsize=6.0,
    maxfontsize=6.0,
)


@contextmanager
def _temp_file_from_file_field(source):
    """
    Create a temp file containing data from a django file field.
    """
    source.open()
    with closing(source):
        try:
            with tempfile.NamedTemporaryFile(delete=False) as destination:
                shutil.copyfileobj(source, destination)
            yield destination.name
        finally:
            os.unlink(destination.name)


def _create_profile(source, get_filename=_temp_file_from_file_field):
    """
    Parse a profile from a django file field source.
    """
    with get_filename(source) as filename:
        return PstatsParser(filename).parse()


def _create_dot(profile, cutoff):
    """
    Create a dot file from pstats data stored in a django file field.
    """
    node_cutoff = cutoff / 100.0
    edge_cutoff = 0.1 / 100.0
    profile.prune(node_cutoff, edge_cutoff, [], False)

    with closing(StringIO()) as fp:
        DotWriter(fp).graph(profile, COLOR_MAP)
        return fp.getvalue()


class ProfileDotView(View):

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, request_id):
        silk_request = get_object_or_404(Request, pk=request_id, prof_file__isnull=False)
        cutoff = float(request.GET.get('cutoff', '') or 5)
        profile = _create_profile(silk_request.prof_file)
        result = dict(dot=_create_dot(profile, cutoff))
        return HttpResponse(json.dumps(result).encode('utf-8'), content_type='application/json')



import base64
import json
import random
import re
from uuid import uuid4

import sqlparse
from django.core.files.storage import get_storage_class
from django.db import models, transaction
from django.db.models import (
    BooleanField,
    CharField,
    DateTimeField,
    FileField,
    FloatField,
    ForeignKey,
    IntegerField,
    ManyToManyField,
    OneToOneField,
    TextField,
)
from django.utils import timezone
from django.utils.safestring import mark_safe

from silk.config import SilkyConfig
from silk.utils.profile_parser import parse_profile

silk_storage = get_storage_class(SilkyConfig().SILKY_STORAGE_CLASS)()


# Seperated out so can use in tests w/o models
def _time_taken(start_time, end_time):
    d = end_time - start_time
    return d.seconds * 1000 + d.microseconds / 1000


def time_taken(self):
    return _time_taken(self.start_time, self.end_time)


class CaseInsensitiveDictionary(dict):
    def __getitem__(self, key):
        return super().__getitem__(key.lower())

    def __setitem__(self, key, value):
        super().__setitem__(key.lower(), value)

    def update(self, other=None, **kwargs):
        for k, v in other.items():
            self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    def __init__(self, d):
        super().__init__()
        for k, v in d.items():
            self[k] = v


class Request(models.Model):
    id = CharField(max_length=36, default=uuid4, primary_key=True)
    path = CharField(max_length=190, db_index=True)
    query_params = TextField(blank=True, default='')
    raw_body = TextField(blank=True, default='')
    body = TextField(blank=True, default='')
    method = CharField(max_length=10)
    start_time = DateTimeField(default=timezone.now, db_index=True)
    view_name = CharField(
        max_length=190, db_index=True, blank=True,
        default='', null=True
    )
    end_time = DateTimeField(null=True, blank=True)
    time_taken = FloatField(blank=True, null=True)
    encoded_headers = TextField(blank=True, default='')  # stores json
    meta_time = FloatField(null=True, blank=True)
    meta_num_queries = IntegerField(null=True, blank=True)
    meta_time_spent_queries = FloatField(null=True, blank=True)
    pyprofile = TextField(blank=True, default='')
    prof_file = FileField(max_length=300, blank=True, storage=silk_storage)

    # Useful method to create shortened copies of strings without losing start and end context
    # Used to ensure path and view_name don't exceed 190 characters
    def _shorten(self, string):
        return f'{string[:94]}...{string[len(string) - 93:]}'

    @property
    def total_meta_time(self):
        return (self.meta_time or 0) + (self.meta_time_spent_queries or 0)

    @property
    def profile_table(self):
        for n, columns in enumerate(parse_profile(self.pyprofile)):
            location = columns[-1]
            if n and '{' not in location and '<' not in location:
                r = re.compile(r'(?P<src>.*\.py)\:(?P<num>[0-9]+).*')
                m = r.search(location)
                group = m.groupdict()
                src = group['src']
                num = group['num']
                name = 'c%d' % n
                fmt = '<a name={name} href="?pos={n}&file_path={src}&line_num={num}#{name}">{location}</a>'
                rep = fmt.format(**dict(group, **locals()))
                yield columns[:-1] + [mark_safe(rep)]
            else:
                yield columns

    # defined in atomic transaction within SQLQuery save()/delete() as well
    # as in bulk_create of SQLQueryManager
    # TODO: This is probably a bad way to do this, .count() will prob do?
    num_sql_queries = IntegerField(default=0)  # TODO replace with count()

    @property
    def time_spent_on_sql_queries(self):
        """
        TODO: Perhaps there is a nicer way to do this with Django aggregates?
        My initial thought was to perform:
        SQLQuery.objects.filter.aggregate(Sum(F('end_time')) - Sum(F('start_time')))
        However this feature isnt available yet, however there has been talk
        for use of F objects within aggregates for four years
        here: https://code.djangoproject.com/ticket/14030. It looks
        like this will go in soon at which point this should be changed.
        """
        return sum(x.time_taken for x in SQLQuery.objects.filter(request=self))

    @property
    def headers(self):
        if self.encoded_headers:
            raw = json.loads(self.encoded_headers)
        else:
            raw = {}

        return CaseInsensitiveDictionary(raw)

    @property
    def content_type(self):
        return self.headers.get('content-type', None)

    @classmethod
    def garbage_collect(cls, force=False):
        """ Remove Request/Responses when we are at the SILKY_MAX_RECORDED_REQUESTS limit
        Note that multiple in-flight requests may call this at once causing a
        double collection """
        check_percent = SilkyConfig().SILKY_MAX_RECORDED_REQUESTS_CHECK_PERCENT
        check_percent /= 100.0
        if check_percent < random.random() and not force:
            return
        target_count = SilkyConfig().SILKY_MAX_RECORDED_REQUESTS

        # Since garbage collection is probabilistic, the target count should
        # be lowered to account for requests before the next garbage collection
        if check_percent != 0:
            target_count -= int(1 / check_percent)

        # Make sure we can delete everything if needed by settings
        if target_count <= 0:
            cls.objects.all().delete()
            return

        try:
            time_cutoff = cls.objects.order_by(
                '-start_time'
            ).values_list(
                'start_time',
                flat=True
            )[target_count]
        except IndexError:
            return

        cls.objects.filter(start_time__lte=time_cutoff).delete()

    def save(self, *args, **kwargs):
        # sometimes django requests return the body as 'None'
        if self.raw_body is None:
            self.raw_body = ''

        if self.body is None:
            self.body = ''

        if self.end_time and self.start_time:
            interval = self.end_time - self.start_time
            self.time_taken = interval.total_seconds() * 1000

        # We can't save if either path or view_name exceed 190 characters
        if self.path and len(self.path) > 190:
            self.path = self._shorten(self.path)

        if self.view_name and len(self.view_name) > 190:
            self.view_name = self._shorten(self.view_name)

        super().save(*args, **kwargs)
        Request.garbage_collect(force=False)


class Response(models.Model):
    id = CharField(max_length=36, default=uuid4, primary_key=True)
    request = OneToOneField(
        Request, related_name='response', db_index=True,
        on_delete=models.CASCADE,
    )
    status_code = IntegerField()
    raw_body = TextField(blank=True, default='')
    body = TextField(blank=True, default='')
    encoded_headers = TextField(blank=True, default='')

    @property
    def content_type(self):
        return self.headers.get('content-type', None)

    @property
    def headers(self):
        if self.encoded_headers:
            raw = json.loads(self.encoded_headers)
        else:
            raw = {}
        return CaseInsensitiveDictionary(raw)

    @property
    def raw_body_decoded(self):
        return base64.b64decode(self.raw_body)


# TODO rewrite docstring
class SQLQueryManager(models.Manager):
    @transaction.atomic
    def bulk_create(self, *args, **kwargs):
        """ensure that num_sql_queries remains consistent. Bulk create does not call
        the model save() method and hence we must add this logic here too"""
        if len(args):
            objs = args[0]
        else:
            objs = kwargs.get('objs')
        for obj in objs:
            obj.prepare_save()

        return super().bulk_create(*args, **kwargs)


class SQLQuery(models.Model):
    query = TextField()
    start_time = DateTimeField(null=True, blank=True, default=timezone.now)
    end_time = DateTimeField(null=True, blank=True)
    time_taken = FloatField(blank=True, null=True)
    identifier = IntegerField(default=-1)
    request = ForeignKey(
        Request, related_name='queries', null=True,
        blank=True, db_index=True, on_delete=models.CASCADE,
    )
    traceback = TextField()
    analysis = TextField(null=True, blank=True)
    objects = SQLQueryManager()

    # TODO docstring
    @property
    def traceback_ln_only(self):
        return '\n'.join(self.traceback.split('\n')[::2])

    @property
    def formatted_query(self):
        return sqlparse.format(self.query, reindent=True, keyword_case='upper')

    @property
    def num_joins(self):
        parsed_query = sqlparse.parse(self.query)
        count = 0
        for statement in parsed_query:
            count += sum(map(lambda t: t.match(sqlparse.tokens.Keyword, r'\.*join\.*', regex=True), statement.flatten()))
        return count

    @property
    def tables_involved(self):
        """
        A really another rudimentary way to work out tables involved in a
        query.
        TODO: Can probably parse the SQL using sqlparse etc and pull out table
        info that way?
        """
        components = [x.strip() for x in self.query.split()]
        tables = []

        for idx, component in enumerate(components):
            # TODO: If django uses aliases on column names they will be falsely
            # identified as tables...
            if component.lower() == 'from' or component.lower() == 'join' or component.lower() == 'as':
                try:
                    _next = components[idx + 1]
                    if not _next.startswith('('):  # Subquery
                        stripped = _next.strip().strip(',')

                        if stripped:
                            tables.append(stripped)
                except IndexError:  # Reach the end
                    pass
        return tables

    def prepare_save(self):
        if self.end_time and self.start_time:
            interval = self.end_time - self.start_time
            self.time_taken = interval.total_seconds() * 1000

        if not self.pk:
            if self.request:
                self.request.num_sql_queries += 1
                self.request.save(update_fields=['num_sql_queries'])

    @transaction.atomic()
    def save(self, *args, **kwargs):
        self.prepare_save()
        super().save(*args, **kwargs)

    @transaction.atomic()
    def delete(self, *args, **kwargs):
        self.request.num_sql_queries -= 1
        self.request.save()
        super().delete(*args, **kwargs)


class BaseProfile(models.Model):
    name = CharField(max_length=300, blank=True, default='')
    start_time = DateTimeField(default=timezone.now)
    end_time = DateTimeField(null=True, blank=True)
    request = ForeignKey(
        Request, null=True, blank=True, db_index=True,
        on_delete=models.CASCADE,
    )
    time_taken = FloatField(blank=True, null=True)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if self.end_time and self.start_time:
            interval = self.end_time - self.start_time
            self.time_taken = interval.total_seconds() * 1000
        super().save(*args, **kwargs)


class Profile(BaseProfile):
    file_path = CharField(max_length=300, blank=True, default='')
    line_num = IntegerField(null=True, blank=True)
    end_line_num = IntegerField(null=True, blank=True)
    func_name = CharField(max_length=300, blank=True, default='')
    exception_raised = BooleanField(default=False)
    queries = ManyToManyField(SQLQuery, related_name='profiles', db_index=True)
    dynamic = BooleanField(default=False)

    @property
    def is_function_profile(self):
        return self.func_name is not None

    @property
    def is_context_profile(self):
        return self.func_name is None

    @property
    def time_spent_on_sql_queries(self):
        return sum(x.time_taken for x in self.queries.all())



from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import NoReverseMatch, reverse

from silk.config import SilkyConfig, default_permissions
from silk.middleware import silky_reverse


class TestAuth(TestCase):
    def test_authentication(self):
        SilkyConfig().SILKY_AUTHENTICATION = True
        response = self.client.get(silky_reverse('requests'))
        self.assertEqual(response.status_code, 302)
        url = response.url
        try:
            # If we run tests within the django_silk project, a login url is available from example_app
            self.assertIn(reverse('login'), url)
        except NoReverseMatch:
            # Otherwise the Django default login url is used, in which case we can test for that instead
            self.assertIn('http://testserver/login/', url)

    def test_default_authorisation(self):
        SilkyConfig().SILKY_AUTHENTICATION = True
        SilkyConfig().SILKY_AUTHORISATION = True
        SilkyConfig().SILKY_PERMISSIONS = default_permissions
        username_and_password = 'bob'  # bob is an imbecile and uses the same pass as his username
        user = User.objects.create(username=username_and_password)
        user.set_password(username_and_password)
        user.save()
        self.client.login(username=username_and_password, password=username_and_password)
        response = self.client.get(silky_reverse('requests'))
        self.assertEqual(response.status_code, 403)
        user.is_staff = True
        user.save()
        response = self.client.get(silky_reverse('requests'))
        self.assertEqual(response.status_code, 200)

    def test_custom_authorisation(self):
        SilkyConfig().SILKY_AUTHENTICATION = True
        SilkyConfig().SILKY_AUTHORISATION = True

        def custom_authorisation(user):
            return user.username.startswith('mike')

        SilkyConfig().SILKY_PERMISSIONS = custom_authorisation
        username_and_password = 'bob'  # bob is an imbecile and uses the same pass as his username
        user = User.objects.create(username=username_and_password)
        user.set_password(username_and_password)
        user.save()
        self.client.login(username=username_and_password, password=username_and_password)
        response = self.client.get(silky_reverse('requests'))
        self.assertEqual(response.status_code, 403)
        user.username = 'mike2'
        user.save()
        response = self.client.get(silky_reverse('requests'))
        self.assertEqual(response.status_code, 200)



from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.contrib.auth import views
from django.urls import include, path

urlpatterns = [
    path(
        route='silk/',
        view=include('silk.urls', namespace='silk'),
    ),
    path(
        route='example_app/',
        view=include('example_app.urls', namespace='example_app'),
    ),
    path(route='admin/', view=admin.site.urls),
]


urlpatterns += [
    path(
        route='login/',
        view=views.LoginView.as_view(
            template_name='example_app/login.html'
        ),
        name='login',
    ),
]


urlpatterns += static(settings.STATIC_URL, document_root=settings.STATIC_ROOT) + \
    static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



# std
import cProfile
import os
import tempfile
from contextlib import contextmanager
from unittest.mock import MagicMock

# 3rd party
from django.test import TestCase
from networkx.drawing.nx_pydot import read_dot

# silk
from silk.views.profile_dot import (
    _create_dot,
    _create_profile,
    _temp_file_from_file_field,
)


class ProfileDotViewTestCase(TestCase):

    @classmethod
    @contextmanager
    def _stats_file(cls):
        """
        Context manager to create some arbitrary profiling stats in a temp file, returning the filename on enter,
        and removing the temp file on exit.
        """
        try:
            with tempfile.NamedTemporaryFile(delete=False) as stats:
                pass
            cProfile.run('1+1', stats.name)
            yield stats.name
        finally:
            os.unlink(stats.name)

    @classmethod
    @contextmanager
    def _stats_data(cls):
        """
        Context manager to create some arbitrary profiling stats in a temp file, returning the data on enter,
        and removing the temp file on exit.
        """
        with cls._stats_file() as filename:
            with open(filename, 'rb') as f:
                yield f.read()

    @classmethod
    def _profile(cls):
        """Create some arbitrary profiling stats."""
        with cls._stats_file() as filename:
            # create profile - we don't need to convert a django file field to a temp file
            # just use the filename of the temp file already created
            @contextmanager
            def dummy(_):
                yield filename
            return _create_profile(filename, dummy)

    @classmethod
    def _mock_file(cls, data):
        """
        Get a mock object that looks like a file but returns data when read is called.
        """
        i = [0]

        def read(n):
            if not i[0]:
                i[0] += 1
                return data

        stream = MagicMock()
        stream.open = lambda: None
        stream.read = read

        return stream

    def test_create_dot(self):
        """
        Verify that a dot file is correctly created from pstats data stored in a file field.
        """
        with self._stats_file():

            try:
                # create dot
                with tempfile.NamedTemporaryFile(delete=False) as dotfile:
                    dot = _create_dot(self._profile(), 5)
                    dotfile.write(dot.encode('utf-8'))

                # verify generated dot is valid
                G = read_dot(dotfile.name)
                self.assertGreater(len(G.nodes()), 0)

            finally:
                os.unlink(dotfile.name)

    def test_temp_file_from_file_field(self):
        """
        Verify that data held in a file like object is copied to a temp file.
        """
        dummy_data = b'dummy data'
        stream = self._mock_file(dummy_data)

        with _temp_file_from_file_field(stream) as filename:
            with open(filename, 'rb') as f:
                self.assertEqual(f.read(), dummy_data)

        # file should have been removed on exit
        self.assertFalse(os.path.exists(filename))



import inspect
import logging
import re
import sys

from silk.profiling.profiler import silk_profile

Logger = logging.getLogger('silk.profiling.dynamic')


def _get_module(module_name):
    """
    Given a module name in form 'path.to.module' return module object for 'module'.
    """
    if '.' in module_name:
        splt = module_name.split('.')
        imp = '.'.join(splt[:-1])
        frm = splt[-1]
        module = __import__(imp, globals(), locals(), [frm], 0)
        module = getattr(module, frm)
    else:
        module = __import__(module_name, globals(), locals(), [], 0)
    return module


def _get_func(module, func_name):
    """
    Given a module and a function name, return the function.

    func_name can be of the forms:
        - 'foo': return a function
        - 'Class.foo': return a method
    """
    cls_name = None
    cls = None
    if '.' in func_name:
        cls_name, func_name = func_name.split('.')
    if cls_name:
        cls = getattr(module, cls_name)
        func = getattr(cls, func_name)
    else:
        func = getattr(module, func_name)
    return cls, func


def profile_function_or_method(module, func, name=None):
    """
    Programmatically apply a decorator to a function in a given module [+ class]

    @param module: module object or module name in form 'path.to.module'
    @param func: function object or function name in form 'foo' or 'Class.method'
    """
    if isinstance(module, str):
        module = _get_module(module)
    decorator = silk_profile(name, _dynamic=True)
    func_name = func
    cls, func = _get_func(module, func_name)
    wrapped_target = decorator(func)
    if cls:
        setattr(cls, func_name.split('.')[-1], wrapped_target)
    else:
        setattr(module, func_name, wrapped_target)


def _get_parent_module(module):
    parent = sys.modules
    splt = module.__name__.split('.')
    if len(splt) > 1:
        for module_name in splt[:-1]:
            try:
                parent = getattr(parent, module_name)
            except AttributeError:
                parent = parent[module_name]
    return parent


def _get_context_manager_source(end_line, file_path, name, start_line):
    inject_code = "with silk_profile('%s', _dynamic=True):\n" % name
    code = 'from silk.profiling.profiler import silk_profile\n'
    with open(file_path) as f:
        ws = ''
        for i, line in enumerate(f):
            if i == start_line:
                # Use the same amount of whitespace as the line currently occupying
                x = re.search(r"^(\s+).*$", line)
                try:
                    ws = x.groups()[0]
                except IndexError:
                    ws = ''
                code += ws + inject_code
                code += ws + '    ' + line
            elif start_line < i <= end_line:
                code += ws + '    ' + line
            else:
                code += line
    return code


def _get_ws(txt):
    """
    Return whitespace at the beginning of a string
    """
    m = re.search(r"^(\s+).*$", txt)
    try:
        fws = m.groups()[0]
    except AttributeError:
        fws = ''
    return fws


def _get_source_lines(func):
    source = inspect.getsourcelines(func)[0]
    fws = _get_ws(source[0])
    for i in range(0, len(source)):
        source[i] = source[i].replace(fws, '', 1)
    return source


def _new_func_from_source(source, func):
    """
    Create new function defined in source but maintain context from func

    @param func: The function whose global + local context we will use
    @param source: Python source code containing def statement
    """
    src_str = ''.join(source)
    frames = inspect.getouterframes(inspect.currentframe())
    calling_frame = frames[2][0]

    context = {}
    # My initial instict was: exec src_str in func.func_globals.items(), calling_frame.f_locals
    # however this seems to break the function closure so caveat here is that we create a new
    # function with the locals merged into the globals.
    #
    # Possible consequences I can think of:
    #   - If a global exists that already has the same name as the local, it will be overwritten in
    #     in the context of this function. This shouldnt matter though as the global should have already
    #     been hidden by the new name?
    #
    # This functionality should be considered experimental as no idea what other consequences there
    # could be.
    #
    # relevant: http://stackoverflow.com/questions/2749655/why-are-closures-broken-within-exec
    globals = func.__globals__
    locals = calling_frame.f_locals
    combined = globals.copy()
    combined.update(locals)
    Logger.debug('New src_str:\n %s' % src_str)
    exec(src_str, combined, context)
    return context[func.__name__]


def _inject_context_manager_func(func, start_line, end_line, name):
    """
    injects a context manager into the given function

    e.g given:

        x = 5
        def foo():
            print x
            print '1'
            print '2'
            print '3'
        inject_context_manager_func(foo, 0, 2, 'cm')

    foo will now have the definition:

        def foo():
            with silk_profile('cm'):
                print x
                print '1'
                print '2'
            print '3'

    closures, globals & locals are honoured

    @param func: object of type<function> or type<instancemethod>
    @param start_line: line at which to inject 'with' statement. line num. is relative to the func, not the module.
    @param end_line: line at which to exit the context
    @param name: name of the profiler
    """
    source = _get_source_lines(func)
    start_line += 1
    end_line += 1
    ws = _get_ws(source[start_line])
    for i in range(start_line, end_line):
        try:
            source[i] = '  ' + source[i]
        except IndexError:
            raise IndexError('Function %s does not have line %d' % (func.__name__, i))

    source.insert(start_line, ws + "from silk.profiling.profiler import silk_profile\n")
    source.insert(start_line + 1, ws + "with silk_profile('%s', _dynamic=True):\n" % name)
    return _new_func_from_source(source, func)


def is_str_typ(o):
    return isinstance(o, str)


def inject_context_manager_func(module, func, start_line, end_line, name):
    if is_str_typ(module):
        module = _get_module(module)
    cls = None
    if is_str_typ(func):
        func_name = func
        cls, func = _get_func(module, func_name)
    else:
        func_name = func.__name__
    new_func = _inject_context_manager_func(func, start_line, end_line, name)
    if cls:
        setattr(cls, func_name, new_func)
    else:
        setattr(module, func_name, new_func)



from django.test import TestCase

from silk import models
from silk.config import SilkyConfig
from silk.middleware import silky_reverse

from .factories import RequestMinFactory


class TestViewClearDB(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        SilkyConfig().SILKY_AUTHENTICATION = False
        SilkyConfig().SILKY_AUTHORISATION = False

    def test_clear_all(self):
        RequestMinFactory.create()
        self.assertEqual(models.Request.objects.count(), 1)
        response = self.client.post(silky_reverse("cleardb"), {"clear_all": "on"})
        self.assertTrue(response.status_code == 200)
        self.assertEqual(models.Request.objects.count(), 0)



import json

from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.code_generation.curl import curl_cmd
from silk.code_generation.django_test_client import gen
from silk.models import Request


class RequestView(View):

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, request_id):
        silk_request = Request.objects.get(pk=request_id)
        query_params = None
        if silk_request.query_params:
            query_params = json.loads(silk_request.query_params)
        body = silk_request.raw_body
        try:
            body = json.loads(body)  # Incase encoded as JSON
        except (ValueError, TypeError):
            pass
        context = {
            'silk_request': silk_request,
            'curl': curl_cmd(url=request.build_absolute_uri(silk_request.path),
                             method=silk_request.method,
                             query_params=query_params,
                             body=body,
                             content_type=silk_request.content_type),
            'query_params': json.dumps(query_params, sort_keys=True, indent=4) if query_params else None,
            'client': gen(path=silk_request.path,
                          method=silk_request.method,
                          query_params=query_params,
                          data=body,
                          content_type=silk_request.content_type),
            'request': request
        }
        return render(request, 'silk/request.html', context)



import random

from django.conf import settings
from django.test import TestCase

from silk.config import SilkyConfig
from silk.middleware import silky_reverse
from silk.views.sql_detail import SQLDetailView

from .test_lib.mock_suite import MockSuite


class TestViewSQLDetail(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        SilkyConfig().SILKY_AUTHENTICATION = False
        SilkyConfig().SILKY_AUTHORISATION = False

    def test_allowed_file_paths_nothing_specified(self):
        """by default we dont display any source, and it should return correctly"""
        request = MockSuite().mock_request()
        query = MockSuite().mock_sql_queries(request=request, n=1)[0]
        response = self.client.get(silky_reverse('request_sql_detail', kwargs={'sql_id': query.id, 'request_id': request.id}))
        self.assertTrue(response.status_code == 200)

    def test_allowed_file_paths_available_source(self):
        """if we request to view source that exists in the TB all should be fine"""
        request = MockSuite().mock_request()
        query = MockSuite().mock_sql_queries(request=request, n=1)[0]
        tb = query.traceback_ln_only
        _, files = SQLDetailView()._urlify(tb)
        file_path = random.choice(files)
        with open(file_path) as f:
            line_num = random.randint(0, len(f.read().split('\n')))
        response = self.client.get(silky_reverse('request_sql_detail',
                                                 kwargs={'sql_id': query.id, 'request_id': request.id}),
                                   data={
                                       'line_num': line_num,
                                       'file_path': file_path
        })
        self.assertTrue(response.status_code == 200)

    def test_allowed_file_paths_unavailable_source(self):
        """if we request to view source that is not in the traceback we should get a 403"""
        request = MockSuite().mock_request()
        query = MockSuite().mock_sql_queries(request=request, n=1)[0]
        file_path = settings.TEMP_DIR + '/blah'
        with open(file_path, 'w') as f:
            f.write('test')
        response = self.client.get(silky_reverse('request_sql_detail',
                                                 kwargs={'sql_id': query.id, 'request_id': request.id}),
                                   data={
                                       'line_num': 0,
                                       'file_path': file_path
        })
        self.assertTrue(response.status_code == 403)



from silk.config import SilkyConfig

__author__ = 'mtford'


def _code(file_path, line_num, end_line_num=None):
    line_num = int(line_num)
    if not end_line_num:
        end_line_num = line_num
    end_line_num = int(end_line_num)
    actual_line = []
    lines = ''
    with open(file_path, encoding='utf-8') as f:
        r = range(max(0, line_num - 10), line_num + 10)
        for i, line in enumerate(f):
            if i in r:
                lines += line
            if i + 1 in range(line_num, end_line_num + 1):
                actual_line.append(line)
    code = lines.split('\n')
    return actual_line, code


def _code_context(file_path, line_num, end_line_num=None, prefix=''):
    actual_line, code = _code(file_path, line_num, end_line_num)
    return {
        prefix + 'code': code,
        prefix + 'file_path': file_path,
        prefix + 'line_num': line_num,
        prefix + 'actual_line': actual_line
    }


def _code_context_from_request(request, end_line_num=None, prefix=''):
    file_path = request.GET.get('file_path')
    line_num = request.GET.get('line_num')
    result = {}
    if file_path is not None and line_num is not None:
        result = _code_context(file_path, line_num, end_line_num, prefix)
    return result


def _should_display_file_name(file_name):
    for ignored_file in SilkyConfig().SILKY_IGNORE_FILES:
        if ignored_file in file_name:
            return False
    return True



import cProfile
import io
import re

import contextlib2 as contextlib
from django.test import TestCase

from silk.utils.profile_parser import parse_profile


class ProfileParserTestCase(TestCase):
    def test_profile_parser(self):
        """
        Verify that the function parse_profile produces the expected output.
        """
        with contextlib.closing(io.StringIO()) as stream:
            with contextlib.redirect_stdout(stream):
                cProfile.run('print()')
            stream.seek(0)
            actual = list(parse_profile(stream))

            # Expected format for the profiling output on cPython implementations (PyPy differs)
            # actual = [
            #     ["ncalls", "tottime", "percall", "cumtime", "percall", "filename:lineno(function)"],
            #     ["1", "0.000", "0.000", "0.000", "0.000", "<string>:1(<module>)"],
            #     ["1", "0.000", "0.000", "0.000", "0.000", "{built-in method builtins.exec}"],
            #     ["1", "0.000", "0.000", "0.000", "0.000", "{built-in method builtins.print}"],
            #     ["1", "0.000", "0.000", "0.000", "0.000", "{method 'disable' of '_lsprof.Profiler' objects}"],
            # ]

            exc_header = ["ncalls", "tottime", "percall", "cumtime", "percall", "filename:lineno(function)"]
            self.assertEqual(actual[0], exc_header)

            exc_number = re.compile(r"\d(.\d+)?")
            exc_module = re.compile(r"({method.*})|({built-in.*})|(<.+>:\d+\(<.+>\))")

            exc_row = [exc_number, exc_number, exc_number, exc_number, exc_number, exc_module]

            for row in actual[1:]:
                for text, expected_regex in zip(row, exc_row):
                    self.assertRegex(
                        text, expected_regex,
                        msg="Expected something like {} but found {}"
                    )



from django.test import TestCase

from silk.views.summary import SummaryView

from .test_lib.mock_suite import MockSuite

mock_suite = MockSuite()


class TestSummaryView(TestCase):
    def test_longest_query_by_view(self):
        [mock_suite.mock_request() for _ in range(0, 10)]
        print([x.time_taken for x in SummaryView()._longest_query_by_view([])])



from django.urls import path

from silk.views.clear_db import ClearDBView
from silk.views.cprofile import CProfileView
from silk.views.profile_detail import ProfilingDetailView
from silk.views.profile_dot import ProfileDotView
from silk.views.profile_download import ProfileDownloadView
from silk.views.profiling import ProfilingView
from silk.views.raw import Raw
from silk.views.request_detail import RequestView
from silk.views.requests import RequestsView
from silk.views.sql import SQLView
from silk.views.sql_detail import SQLDetailView
from silk.views.summary import SummaryView

app_name = 'silk'
urlpatterns = [
    path(route='', view=SummaryView.as_view(), name='summary'),
    path(route='requests/', view=RequestsView.as_view(), name='requests'),
    path(
        route='request/<uuid:request_id>/',
        view=RequestView.as_view(),
        name='request_detail',
    ),
    path(
        route='request/<uuid:request_id>/sql/',
        view=SQLView.as_view(),
        name='request_sql',
    ),
    path(
        route='request/<uuid:request_id>/sql/<int:sql_id>/',
        view=SQLDetailView.as_view(),
        name='request_sql_detail',
    ),
    path(
        route='request/<uuid:request_id>/raw/',
        view=Raw.as_view(),
        name='raw',
    ),
    path(
        route='request/<uuid:request_id>/pyprofile/',
        view=ProfileDownloadView.as_view(),
        name='request_profile_download',
    ),
    path(
        route='request/<uuid:request_id>/json/',
        view=ProfileDotView.as_view(),
        name='request_profile_dot',
    ),
    path(
        route='request/<uuid:request_id>/profiling/',
        view=ProfilingView.as_view(),
        name='request_profiling',
    ),
    path(
        route='request/<uuid:request_id>/profile/<int:profile_id>/',
        view=ProfilingDetailView.as_view(),
        name='request_profile_detail',
    ),
    path(
        route='request/<uuid:request_id>/profile/<int:profile_id>/sql/',
        view=SQLView.as_view(),
        name='request_and_profile_sql',
    ),
    path(
        route='request/<uuid:request_id>/profile/<int:profile_id>/sql/<int:sql_id>/',
        view=SQLDetailView.as_view(),
        name='request_and_profile_sql_detail',
    ),
    path(
        route='profile/<int:profile_id>/',
        view=ProfilingDetailView.as_view(),
        name='profile_detail',
    ),
    path(
        route='profile/<int:profile_id>/sql/',
        view=SQLView.as_view(),
        name='profile_sql',
    ),
    path(
        route='profile/<int:profile_id>/sql/<int:sql_id>/',
        view=SQLDetailView.as_view(),
        name='profile_sql_detail',
    ),
    path(route='profiling/', view=ProfilingView.as_view(), name='profiling'),
    path(route='cleardb/', view=ClearDBView.as_view(), name='cleardb'),
    path(
        route='request/<uuid:request_id>/cprofile/',
        view=CProfileView.as_view(),
        name='cprofile',
    ),
]



import random
import unittest
from unittest.mock import Mock

from django.test import TestCase

from silk.views.requests import RequestsView

from .test_lib.assertion import dict_contains
from .test_lib.mock_suite import MockSuite


class TestRootViewDefaults(TestCase):
    def test_path(self):
        requests = [MockSuite().mock_request() for _ in range(0, 3)]
        paths = RequestsView()._get_paths()
        for r in requests:
            self.assertIn(r.path, paths)

    def test_show(self):
        self.assertIn(RequestsView.default_show, RequestsView.show)

    def test_order_by(self):
        self.assertIn(RequestsView.default_order_by, RequestsView.order_by)


class TestContext(TestCase):
    def test_default(self):
        request = Mock(spec_set=['GET', 'session'])
        request.session = {}
        request.GET = {}
        context = RequestsView()._create_context(request)
        self.assertTrue(dict_contains({
            'show': RequestsView.default_show,
            'order_by': RequestsView.default_order_by,
            'options_show': RequestsView.show,
            'options_order_by': RequestsView().options_order_by,
            'options_order_dir': RequestsView().options_order_dir,
        }, context))
        self.assertQuerysetEqual(context['options_paths'], RequestsView()._get_paths())
        self.assertNotIn('path', context)
        self.assertIn('results', context)

    def test_get(self):
        request = Mock(spec_set=['GET', 'session'])
        request.session = {}
        show = 10
        path = '/path/to/somewhere/'
        order_by = 'path'
        request.GET = {'show': show,
                       'path': path,
                       'order_by': order_by}
        context = RequestsView()._create_context(request)
        self.assertTrue(dict_contains({
            'show': show,
            'order_by': order_by,
            'path': path,
            'options_show': RequestsView.show,
            'options_order_by': RequestsView().options_order_by,
            'options_order_dir': RequestsView().options_order_dir,
        }, context))
        self.assertQuerysetEqual(context['options_paths'], RequestsView()._get_paths())
        self.assertIn('results', context)


class TestGetObjects(TestCase):
    def assertSorted(self, objects, sort_field):
        for idx, r in enumerate(objects):
            try:
                nxt = objects[idx + 1]
                self.assertGreaterEqual(getattr(r, sort_field), getattr(nxt, sort_field))
            except IndexError:
                pass

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.requests = [MockSuite().mock_request() for _ in range(0, 50)]

    def test_defaults(self):
        objects = RequestsView()._get_objects()
        self.assertEqual(len(objects), 25)
        self.assertSorted(objects, 'start_time')

    def test_show(self):
        objects = RequestsView()._get_objects(show=10)
        self.assertEqual(len(objects), 10)

    def test_path(self):
        request = random.choice(self.requests)
        objects = RequestsView()._get_objects(path=request.path)
        for r in objects:
            self.assertEqual(r.path, request.path)

    @unittest.skip("Flaky")
    def test_time_spent_db_with_path(self):
        request = random.choice(self.requests)
        query_set = RequestsView()._get_objects(order_by='db_time',
                                                path=request.path)
        num_results = query_set.count()
        self.assertTrue(num_results)
        for result in query_set:
            self.assertEqual(result.path, request.path)


class TestOrderingRequestView(TestCase):
    def assertSorted(self, objects, sort_field):
        for idx, r in enumerate(objects):
            try:
                nxt = objects[idx + 1]
                self.assertGreaterEqual(getattr(r, sort_field), getattr(nxt, sort_field))
            except IndexError:
                pass

    def test_ordering(self):
        self.assertSorted(objects=RequestsView()._get_objects(order_by='start_time'),
                          sort_field='start_time')
        self.assertSorted(objects=RequestsView()._get_objects(order_by='path'),
                          sort_field='path')
        self.assertSorted(objects=RequestsView()._get_objects(order_by='num_sql_queries'),
                          sort_field='num_sql_queries')
        self.assertSorted(objects=RequestsView()._get_objects(order_by='time_taken'),
                          sort_field='time_taken')
        self.assertSorted(objects=RequestsView()._get_objects(order_by='db_time'),
                          sort_field='db_time')



from django.db import transaction
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.models import Profile, Request, Response, SQLQuery
from silk.utils.data_deletion import delete_model


@method_decorator(transaction.non_atomic_requests, name="dispatch")
class ClearDBView(View):

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, *_, **kwargs):
        return render(request, 'silk/clear_db.html')

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def post(self, request, *_, **kwargs):
        context = {}
        if 'clear_all' in request.POST:
            delete_model(Profile)
            delete_model(SQLQuery)
            delete_model(Response)
            delete_model(Request)
            tables = ['Response', 'SQLQuery', 'Profile', 'Request']
            context['msg'] = 'Cleared data for following silk tables: {}'.format(', '.join(tables))
        return render(request, 'silk/clear_db.html', context=context)
